"""Test methods for cg cli set sample"""

import pytest
from click.testing import CliRunner

from cg.cli.set.base import sample
from cg.constants import EXIT_SUCCESS, Priority
from cg.constants.subject import Sex
from cg.models.cg_config import CGConfig
from cg.store.models import Sample
from cg.store.store import Store


def test_invalid_sample(cli_runner: CliRunner, base_context: CGConfig):
    # GIVEN an empty database

    # WHEN running set with a sample that does not exist
    sample_id = "dummy_sample_id"
    result = cli_runner.invoke(sample, [sample_id], obj=base_context)

    # THEN it should complain on invalid sample
    assert result.exit_code != EXIT_SUCCESS


def test_skip_lims(cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers):
    # GIVEN a database with a sample

    db_sample = helpers.add_sample(base_store, sex=Sex.FEMALE)
    key = "name"
    new_value = "new_value"

    # WHEN setting sample but skipping lims
    result = cli_runner.invoke(
        sample,
        [db_sample.internal_id, "-kv", "name", "dummy_value", "-y", "--skip-lims"],
        obj=base_context,
    )

    # THEN update sample should have no recorded update key and value
    assert result.exit_code == EXIT_SUCCESS
    assert base_context.lims_api.get_updated_sample_key() != key
    assert base_context.lims_api.get_updated_sample_value() != new_value


@pytest.mark.parametrize("key", ["name", "capture_kit"])
def test_set_sample(cli_runner: CliRunner, base_context: CGConfig, base_store: Store, key, helpers):
    # GIVEN a database with a sample

    db_sample = helpers.add_sample(base_store, sex=Sex.FEMALE)
    new_value = "new_value"
    assert getattr(db_sample, key) != new_value

    # WHEN setting key on sample to new_value
    result = cli_runner.invoke(
        sample,
        [db_sample.internal_id, "-kv", key, new_value, "-y"],
        obj=base_context,
    )

    # THEN it should have new_value as attribute key on the sample and in LIMS
    assert result.exit_code == EXIT_SUCCESS
    assert getattr(db_sample, key) == new_value
    assert base_context.lims_api.get_updated_sample_key() == key
    assert base_context.lims_api.get_updated_sample_value() == new_value


@pytest.mark.parametrize("new_value", ["false", "true", "True", "False"])
def test_set_boolean_sample(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, new_value, helpers
):
    # GIVEN a database with a sample

    db_sample = helpers.add_sample(base_store, sex=Sex.FEMALE)
    value: bool = new_value.lower() == "true"

    # WHEN setting key on sample to new_value
    result = cli_runner.invoke(
        sample,
        [db_sample.internal_id, "-kv", "no_invoice", new_value, "-y", "--skip-lims"],
        obj=base_context,
    )

    # THEN it should have new_value as attribute key on the sample
    assert result.exit_code == EXIT_SUCCESS
    assert getattr(db_sample, "no_invoice") == value


def test_sex(cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers):
    # GIVEN a database with a sample

    db_sample = helpers.add_sample(base_store, sex=Sex.FEMALE)
    key = "sex"
    new_value = "male"
    assert getattr(db_sample, key) != new_value

    # WHEN setting key on sample to new_value
    result = cli_runner.invoke(
        sample, [db_sample.internal_id, "-kv", key, new_value, "-y"], obj=base_context
    )

    # THEN it should have new_value as attribute key on the sample and in LIMS
    assert result.exit_code == EXIT_SUCCESS
    assert getattr(db_sample, key) == new_value
    assert base_context.lims_api.get_updated_sample_key() == key
    assert base_context.lims_api.get_updated_sample_value() == new_value


def test_priority_text(cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers):
    # GIVEN a database with a sample
    db_sample = helpers.add_sample(base_store, sex=Sex.FEMALE)
    key = "priority"
    new_value = Priority.express.name
    assert db_sample.priority_human != new_value

    # WHEN setting key on sample to new_value
    result = cli_runner.invoke(
        sample, [db_sample.internal_id, "-kv", key, new_value, "-y"], obj=base_context
    )

    # THEN it should have new_value as attribute key on the sample and in LIMS
    assert result.exit_code == EXIT_SUCCESS
    assert db_sample.priority_human == new_value
    assert base_context.lims_api.get_updated_sample_key() == key
    assert base_context.lims_api.get_updated_sample_value() == db_sample.priority_human


def test_priority_number(cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers):
    # GIVEN a database with a sample
    db_sample = helpers.add_sample(base_store, sex=Sex.FEMALE)
    key = "priority"
    new_value = Priority.express
    assert db_sample.priority != new_value

    # WHEN setting key on sample to new_value
    result = cli_runner.invoke(
        sample,
        [db_sample.internal_id, "-kv", key, new_value.name, "-y"],
        obj=base_context,
    )

    # THEN it should have new_value as attribute key on the sample and in LIMS
    assert result.exit_code == EXIT_SUCCESS
    assert db_sample.priority == new_value
    assert base_context.lims_api.get_updated_sample_key() == key
    assert base_context.lims_api.get_updated_sample_value() == db_sample.priority_human


def test_sample_comment(cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers):
    # GIVEN a database with a sample without comment
    db_sample = helpers.add_sample(base_store, sex=Sex.FEMALE)
    key = "comment"

    # WHEN setting key on sample to new_value
    new_value = "test comment"
    result = cli_runner.invoke(
        sample, [db_sample.internal_id, "-kv", key, new_value, "-y"], obj=base_context
    )

    # THEN it should have new_value as attribute key on the sample and in LIMS
    assert result.exit_code == EXIT_SUCCESS
    assert getattr(db_sample, key).endswith(new_value)
    assert base_context.lims_api.get_updated_sample_key() == key
    assert base_context.lims_api.get_updated_sample_value() == new_value


def test_sample_comment_append(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers
):
    # GIVEN a database with a sample with a comment
    old_value = "test comment"
    sample_obj = helpers.add_sample(
        base_store, sex=Sex.FEMALE, comment=f"2022-06-13 10:16-test.user: {old_value}"
    )
    key = "comment"

    # WHEN setting key again on sample this time to new_value
    new_value = "another test comment"
    result = cli_runner.invoke(
        sample, [sample_obj.internal_id, "-kv", key, new_value, "-y"], obj=base_context
    )

    # THEN it should have new_value above old_value as attribute key on the sample and in LIMS
    comments = getattr(sample_obj, key).split("\n")
    assert result.exit_code == EXIT_SUCCESS
    assert comments[0].endswith(new_value)
    assert comments[1].endswith(old_value)
    assert base_context.lims_api.get_updated_sample_key() == key
    assert base_context.lims_api.get_updated_sample_value() == new_value


def test_invalid_customer(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers
):
    # GIVEN a database with a sample
    sample_id = helpers.add_sample(base_store).internal_id
    customer_id = "dummy_customer_id"
    sample_query = base_store._get_query(table=Sample)

    assert sample_query.first().customer.internal_id != customer_id

    # WHEN calling set sample with an invalid customer
    result = cli_runner.invoke(
        sample, [sample_id, "-kv", "customer", customer_id, "-y", "--skip-lims"], obj=base_context
    )

    # THEN it should error about missing customer instead of setting the value
    assert result.exit_code != EXIT_SUCCESS
    assert sample_query.first().customer.internal_id != customer_id


def test_customer(cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers):
    # GIVEN a database with a sample and two customers
    sample_id = helpers.add_sample(base_store).internal_id
    customer_id = helpers.ensure_customer(base_store, "another_customer").internal_id
    sample_query = base_store._get_query(table=Sample)

    assert sample_query.first().customer.internal_id != customer_id

    # WHEN calling set sample with a valid customer
    result = cli_runner.invoke(
        sample, [sample_id, "-kv", "customer", customer_id, "-y", "--skip-lims"], obj=base_context
    )

    # THEN it should set the customer of the sample
    assert result.exit_code == EXIT_SUCCESS
    assert sample_query.first().customer.internal_id == customer_id


def test_invalid_downsampled_to(cli_runner: CliRunner, base_context: CGConfig):
    # GIVEN a database with a sample
    downsampled_to = "downsampled_to"

    # WHEN calling set sample with an invalid value of downsampled to
    result = cli_runner.invoke(
        sample, ["dummy_sample_id", "-kv", "downsampled_to", downsampled_to, "-y"], obj=base_context
    )

    # THEN wrong data type
    assert result.exit_code != EXIT_SUCCESS


def test_downsampled_to(cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers):
    # GIVEN a database with a sample
    sample_id = helpers.add_sample(base_store).internal_id
    downsampled_to = 111111
    sample_query = base_store._get_query(table=Sample)

    assert sample_query.first().downsampled_to != downsampled_to

    # WHEN calling set sample with a valid value of downsampled to
    result = cli_runner.invoke(
        sample, [sample_id, "-kv", "downsampled_to", downsampled_to, "-y"], obj=base_context
    )

    # THEN the value should have been set on the sample
    assert result.exit_code == EXIT_SUCCESS
    assert sample_query.first().downsampled_to == downsampled_to


def test_reset_downsampled_to(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers
):
    # GIVEN a database with a sample
    sample_id = helpers.add_sample(base_store).internal_id
    downsampled_to = 0
    sample_query = base_store._get_query(table=Sample)

    assert sample_query.first().downsampled_to != downsampled_to

    # WHEN calling set sample with a valid reset value of downsampled to
    result = cli_runner.invoke(
        sample, [sample_id, "-kv", "downsampled_to", "", "-y"], obj=base_context
    )

    # THEN the value should have been set on the sample
    assert result.exit_code == EXIT_SUCCESS
    assert not sample_query.first().downsampled_to


def test_invalid_application(
    cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers
):
    # GIVEN a database with a sample
    sample_id = helpers.add_sample(base_store).internal_id
    application_tag = "dummy_application"
    sample_query = base_store._get_query(table=Sample)

    assert sample_query.first().application_version.application.tag != application_tag

    # WHEN calling set sample with an invalid application
    result = cli_runner.invoke(
        sample,
        [sample_id, "-kv", "application_version", application_tag, "-y", "--skip-lims"],
        obj=base_context,
    )

    # THEN it should error about missing application instead of setting the value
    assert result.exit_code != EXIT_SUCCESS
    assert sample_query.first().application_version.application.tag != application_tag


def test_application(cli_runner: CliRunner, base_context: CGConfig, base_store: Store, helpers):
    # GIVEN a database with a sample and two applications
    sample_obj = helpers.add_sample(base_store)
    application_tag = helpers.ensure_application_version(
        base_store, "another_application"
    ).application.tag
    sample_query = base_store._get_query(table=Sample)

    assert sample_query.first().application_version.application.tag != application_tag

    # WHEN calling set sample with an invalid application
    result = cli_runner.invoke(
        sample,
        [
            sample_obj.internal_id,
            "-kv",
            "application_version",
            application_tag,
            "-y",
        ],
        obj=base_context,
    )

    # THEN the application should have been set in status db
    assert result.exit_code == EXIT_SUCCESS
    assert sample_obj.application_version.application.tag == application_tag
    # THEN the application should have been set in LIMS
    assert base_context.lims_api.get_updated_sample_key() == "application"
    assert base_context.lims_api.get_updated_sample_value() == application_tag



import logging
from datetime import datetime
from pathlib import Path
from typing import Any

from cg.constants import Workflow
from cg.constants.constants import FileExtensions, FileFormat, MultiQC, WorkflowManager
from cg.constants.nextflow import NFX_WORK_DIR
from cg.constants.tb import AnalysisStatus
from cg.exc import CgError, MetricsQCError
from cg.io.controller import ReadFile, WriteFile
from cg.io.yaml import write_yaml_nextflow_style
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.nf_handlers import NextflowHandler, NfTowerHandler
from cg.models.cg_config import CGConfig
from cg.models.deliverables.metric_deliverables import (
    MetricsBase,
    MetricsDeliverablesCondition,
)
from cg.models.fastq import FastqFileMeta
from cg.models.nf_analysis import FileDeliverable, PipelineDeliverables
from cg.models.rnafusion.rnafusion import CommandArgs
from cg.utils import Process

LOG = logging.getLogger(__name__)


class NfAnalysisAPI(AnalysisAPI):
    """Parent class for handling NF-core analyses."""

    def __init__(self, config: CGConfig, workflow: Workflow):
        super().__init__(workflow=workflow, config=config)
        self.workflow: Workflow = workflow
        self.root_dir: str | None = None
        self.nfcore_workflow_path: str | None = None
        self.references: str | None = None
        self.profile: str | None = None
        self.conda_env: str | None = None
        self.conda_binary: str | None = None
        self.tower_binary_path: str | None = None
        self.tower_workflow: str | None = None
        self.account: str | None = None
        self.email: str | None = None
        self.compute_env_base: str | None = None
        self.revision: str | None = None
        self.nextflow_binary_path: str | None = None

    @property
    def root(self) -> str:
        return self.root_dir

    @property
    def process(self):
        if not self._process:
            self._process = Process(
                binary=self.tower_binary_path,
            )
        return self._process

    @process.setter
    def process(self, process: Process):
        self._process = process

    def get_profile(self, profile: str | None = None) -> str:
        """Get NF profiles."""
        return profile or self.profile

    def get_workflow_manager(self) -> str:
        """Get workflow manager from Tower."""
        return WorkflowManager.Tower.value

    def get_workflow_version(self, case_id: str) -> str:
        """Get pipeline version from config."""
        return self.revision

    def get_case_path(self, case_id: str) -> Path:
        """Path to case working directory."""
        return Path(self.root_dir, case_id)

    def get_sample_sheet_path(self, case_id: str) -> Path:
        """Path to sample sheet."""
        return Path(self.get_case_path(case_id), f"{case_id}_samplesheet").with_suffix(
            FileExtensions.CSV
        )

    def get_compute_env(self, case_id: str) -> str:
        """Get the compute environment for the head job based on the case priority."""
        return f"{self.compute_env_base}-{self.get_slurm_qos_for_case(case_id=case_id)}"

    @staticmethod
    def get_nextflow_config_path(nextflow_config: str | None = None) -> Path | None:
        """Path to Nextflow config file."""
        if nextflow_config:
            return Path(nextflow_config).absolute()

    def get_job_ids_path(self, case_id: str) -> Path:
        """Return the path to a Trailblazer config file containing Tower IDs."""
        return Path(self.root_dir, case_id, "tower_ids").with_suffix(FileExtensions.YAML)

    def get_deliverables_file_path(self, case_id: str) -> Path:
        """Path to deliverables file for a case."""
        return Path(self.get_case_path(case_id), f"{case_id}_deliverables").with_suffix(
            FileExtensions.YAML
        )

    def get_metrics_deliverables_path(self, case_id: str) -> Path:
        """Return a path where the <case>_metrics_deliverables.yaml file should be located."""
        return Path(self.root_dir, case_id, f"{case_id}_metrics_deliverables").with_suffix(
            FileExtensions.YAML
        )

    def get_params_file_path(self, case_id: str, params_file: Path | None = None) -> Path:
        """Return parameters file or a path where the default parameters file for a case id should be located."""
        if params_file:
            return Path(params_file).absolute()
        return Path((self.get_case_path(case_id)), f"{case_id}_params_file").with_suffix(
            FileExtensions.YAML
        )

    def create_case_directory(self, case_id: str, dry_run: bool = False) -> None:
        """Create case directory."""
        if not dry_run:
            Path(self.get_case_path(case_id=case_id)).mkdir(parents=True, exist_ok=True)

    def get_log_path(self, case_id: str, workflow: str, log: str = None) -> Path:
        """Path to NF log."""
        if log:
            return log
        launch_time: str = datetime.now().strftime("%Y-%m-%d_%H.%M.%S")
        return Path(
            self.get_case_path(case_id),
            f"{case_id}_{workflow}_nextflow_{launch_time}",
        ).with_suffix(FileExtensions.LOG)

    def get_workdir_path(self, case_id: str, work_dir: Path | None = None) -> Path:
        """Path to NF work directory."""
        if work_dir:
            return work_dir.absolute()
        return Path(self.get_case_path(case_id), NFX_WORK_DIR)

    @staticmethod
    def extract_read_files(
        metadata: list[FastqFileMeta], forward_read: bool = False, reverse_read: bool = False
    ) -> list[str]:
        """Extract a list of fastq file paths for either forward or reverse reads."""
        if forward_read and not reverse_read:
            read_direction = 1
        elif reverse_read and not forward_read:
            read_direction = 2
        else:
            raise ValueError("Either forward or reverse needs to be specified")
        sorted_metadata: list = sorted(metadata, key=lambda k: k.path)
        return [
            fastq_file.path
            for fastq_file in sorted_metadata
            if fastq_file.read_direction == read_direction
        ]

    def verify_sample_sheet_exists(self, case_id: str, dry_run: bool = False) -> None:
        """Raise an error if sample sheet file is not found."""
        if not dry_run and not Path(self.get_sample_sheet_path(case_id=case_id)).exists():
            raise ValueError(f"No config file found for case {case_id}")

    def verify_deliverables_file_exists(self, case_id: str) -> None:
        """Raise an error if a deliverable file is not found."""
        if not Path(self.get_deliverables_file_path(case_id=case_id)).exists():
            raise CgError(f"No deliverables file found for case {case_id}")

    def write_params_file(self, case_id: str, workflow_parameters: dict) -> None:
        """Write params-file for analysis."""
        LOG.debug("Writing parameters file")
        write_yaml_nextflow_style(
            content=workflow_parameters,
            file_path=self.get_params_file_path(case_id=case_id),
        )

    @staticmethod
    def write_sample_sheet(
        content: list[list[Any]],
        file_path: Path,
        header: list[str],
    ) -> None:
        """Write sample sheet CSV file."""
        LOG.debug("Writing sample sheet")
        if header:
            content.insert(0, header)
        WriteFile.write_file_from_content(
            content=content,
            file_format=FileFormat.CSV,
            file_path=file_path,
        )

    @staticmethod
    def write_deliverables_file(
        deliverables_content: dict, file_path: Path, file_format=FileFormat.YAML
    ) -> None:
        """Write deliverables file."""
        WriteFile.write_file_from_content(
            content=deliverables_content, file_format=file_format, file_path=file_path
        )

    def write_trailblazer_config(self, case_id: str, tower_id: str) -> None:
        """Write Tower IDs to a file used as the Trailblazer config."""
        config_path: Path = self.get_job_ids_path(case_id=case_id)
        LOG.info(f"Writing Tower ID to {config_path.as_posix()}")
        WriteFile.write_file_from_content(
            content={case_id: [tower_id]},
            file_format=FileFormat.YAML,
            file_path=config_path,
        )

    def _run_analysis_with_nextflow(
        self, case_id: str, command_args: CommandArgs, dry_run: bool
    ) -> None:
        """Run analysis with given options using Nextflow."""
        self.process = Process(
            binary=self.nextflow_binary_path,
            environment=self.conda_env,
            conda_binary=self.conda_binary,
            launch_directory=self.get_case_path(case_id=case_id),
        )
        LOG.info("Workflow will be executed using Nextflow")
        parameters: list[str] = NextflowHandler.get_nextflow_run_parameters(
            case_id=case_id,
            pipeline_path=self.nfcore_workflow_path,
            root_dir=self.root_dir,
            command_args=command_args.dict(),
        )
        self.process.export_variables(
            export=NextflowHandler.get_variables_to_export(),
        )
        command: str = self.process.get_command(parameters=parameters)
        LOG.info(f"{command}")
        sbatch_number: int = NextflowHandler.execute_head_job(
            case_id=case_id,
            case_directory=self.get_case_path(case_id=case_id),
            slurm_account=self.account,
            email=self.email,
            qos=self.get_slurm_qos_for_case(case_id=case_id),
            commands=command,
            dry_run=dry_run,
        )
        LOG.info(f"Nextflow head job running as job: {sbatch_number}")

    def _run_analysis_with_tower(
        self, case_id: str, command_args: CommandArgs, dry_run: bool
    ) -> None:
        """Run analysis with given options using NF-Tower."""
        LOG.info("Workflow will be executed using Tower")
        if command_args.resume:
            from_tower_id: int = command_args.id or NfTowerHandler.get_last_tower_id(
                case_id=case_id,
                trailblazer_config=self.get_job_ids_path(case_id=case_id),
            )
            LOG.info(f"Workflow will be resumed from run with Tower id: {from_tower_id}.")
            parameters: list[str] = NfTowerHandler.get_tower_relaunch_parameters(
                from_tower_id=from_tower_id, command_args=command_args.dict()
            )
        else:
            parameters: list[str] = NfTowerHandler.get_tower_launch_parameters(
                tower_workflow=self.tower_workflow, command_args=command_args.dict()
            )
        self.process.run_command(parameters=parameters, dry_run=dry_run)
        if self.process.stderr:
            LOG.error(self.process.stderr)
        if not dry_run:
            tower_id = NfTowerHandler.get_tower_id(stdout_lines=self.process.stdout_lines())
            self.write_trailblazer_config(case_id=case_id, tower_id=tower_id)
        LOG.info(self.process.stdout)

    def run_analysis(
        self,
        case_id: str,
        command_args: CommandArgs,
        use_nextflow: bool,
        dry_run: bool = False,
    ) -> None:
        """Execute run analysis with given options."""
        if use_nextflow:
            self._run_analysis_with_nextflow(
                case_id=case_id,
                command_args=command_args,
                dry_run=dry_run,
            )
        else:
            self._run_analysis_with_tower(
                case_id=case_id,
                command_args=command_args,
                dry_run=dry_run,
            )

    @staticmethod
    def get_deliverables_template_content() -> list[dict]:
        """Return deliverables file template content."""
        raise NotImplementedError

    def get_deliverables_for_case(self, case_id: str) -> PipelineDeliverables:
        """Return PipelineDeliverables for a given case."""
        deliverable_template: list[dict] = self.get_deliverables_template_content()
        sample_id: str = self.status_db.get_samples_by_case_id(case_id).pop().internal_id
        files: list[FileDeliverable] = []
        for file in deliverable_template:
            for deliverable_field, deliverable_value in file.items():
                if deliverable_value is None:
                    continue
                file[deliverable_field] = file[deliverable_field].replace("CASEID", case_id)
                file[deliverable_field] = file[deliverable_field].replace("SAMPLEID", sample_id)
                file[deliverable_field] = file[deliverable_field].replace(
                    "PATHTOCASE", str(self.get_case_path(case_id=case_id))
                )
            files.append(FileDeliverable(**file))
        return PipelineDeliverables(files=files)

    def get_multiqc_json_path(self, case_id: str) -> Path:
        """Return the path of the multiqc_data.json file."""
        return Path(
            self.root_dir,
            case_id,
            MultiQC.MULTIQC,
            MultiQC.MULTIQC_DATA,
            MultiQC.MULTIQC_DATA + FileExtensions.JSON,
        )

    def get_workflow_metrics(self) -> dict:
        """Get nf-core pipeline metrics constants."""
        return {}

    def get_multiqc_json_metrics(self, case_id: str) -> list[MetricsBase]:
        """Return a list of the metrics specified in a MultiQC json file."""
        raise NotImplementedError

    def get_metric_base_list(self, sample_id: str, metrics_values: dict) -> list[MetricsBase]:
        """Return a list of MetricsBase objects for a given sample."""
        metric_base_list: list[MetricsBase] = []
        for metric_name, metric_value in metrics_values.items():
            metric_base_list.append(
                MetricsBase(
                    header=None,
                    id=sample_id,
                    input=MultiQC.MULTIQC_DATA + FileExtensions.JSON,
                    name=metric_name,
                    step=MultiQC.MULTIQC,
                    value=metric_value,
                    condition=self.get_workflow_metrics().get(metric_name, None),
                )
            )
        return metric_base_list

    @staticmethod
    def ensure_mandatory_metrics_present(metrics: list[MetricsBase]) -> None:
        return None

    def create_metrics_deliverables_content(self, case_id: str) -> dict[str, list[dict[str, Any]]]:
        """Create the content of metrics deliverables file."""
        metrics: list[MetricsBase] = self.get_multiqc_json_metrics(case_id=case_id)
        self.ensure_mandatory_metrics_present(metrics=metrics)
        return {"metrics": [metric.dict() for metric in metrics]}

    def write_metrics_deliverables(self, case_id: str, dry_run: bool = False) -> None:
        """Write <case>_metrics_deliverables.yaml file."""
        metrics_deliverables_path: Path = self.get_metrics_deliverables_path(case_id=case_id)
        content: dict = self.create_metrics_deliverables_content(case_id=case_id)
        if dry_run:
            LOG.info(
                f"Dry-run: metrics deliverables file would be written to {metrics_deliverables_path.as_posix()}"
            )
            return

        LOG.info(f"Writing metrics deliverables file to {metrics_deliverables_path.as_posix()}")
        WriteFile.write_file_from_content(
            content=content,
            file_format=FileFormat.YAML,
            file_path=metrics_deliverables_path,
        )

    def validate_qc_metrics(self, case_id: str, dry_run: bool = False) -> None:
        """Validate the information from a QC metrics deliverable file."""

        if dry_run:
            LOG.info("Dry-run: QC metrics validation would be performed")
            return

        LOG.info("Validating QC metrics")
        try:
            metrics_deliverables_path: Path = self.get_metrics_deliverables_path(case_id=case_id)
            qc_metrics_raw: dict = ReadFile.get_content_from_file(
                file_format=FileFormat.YAML, file_path=metrics_deliverables_path
            )
            MetricsDeliverablesCondition(**qc_metrics_raw)
        except MetricsQCError as error:
            LOG.error(f"QC metrics failed for {case_id}")
            self.trailblazer_api.set_analysis_status(case_id=case_id, status=AnalysisStatus.FAILED)
            self.trailblazer_api.add_comment(case_id=case_id, comment=str(error))
            raise MetricsQCError from error
        except CgError as error:
            LOG.error(f"Could not create metrics deliverables file: {error}")
            self.trailblazer_api.set_analysis_status(case_id=case_id, status=AnalysisStatus.ERROR)
            raise CgError from error
        self.trailblazer_api.set_analysis_status(case_id=case_id, status=AnalysisStatus.COMPLETED)



import click

from cg.constants.sequencing import SequencingPlatform

OPTION_FROM_START = click.option(
    "--from-start",
    is_flag=True,
    default=False,
    show_default=True,
    help="Start workflow from the start",
)

OPTION_INSTRUMENT_PLATFORM = click.option(
    "--instrument-platform",
    show_default=True,
    default=SequencingPlatform.ILLUMINA,
    type=click.Choice([SequencingPlatform.ILLUMINA, SequencingPlatform.OXFORD_NANOPORE]),
    help="Instrument platform. Set this option to override the default.",
)



"""Path fixtures for demultiplex tests."""

import shutil
from pathlib import Path

import pytest

from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.constants.nanopore_files import NanoporeDirsAndFiles
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData

CORRECT_SAMPLE_SHEET: str = "CorrectSampleSheet.csv"


@pytest.fixture
def tmp_flow_cells_directory(tmp_path: Path, flow_cells_dir: Path) -> Path:
    """
    Return the path to a temporary flow cells directory with flow cells ready for demultiplexing.
    Generates a copy of the original flow cells directory
    """
    original_dir = flow_cells_dir
    tmp_dir = Path(tmp_path, "flow_cells")

    return Path(shutil.copytree(original_dir, tmp_dir))


@pytest.fixture
def tmp_broken_flow_cells_directory(tmp_path: Path, broken_flow_cells_dir: Path) -> Path:
    """
    Return the path to a temporary flow cells directory with incomplete or broken flow cells.
    Generates a copy of the original flow cells directory
    """
    original_dir = broken_flow_cells_dir
    tmp_dir = Path(tmp_path, "broken_flow_cells")

    return Path(shutil.copytree(original_dir, tmp_dir))


@pytest.fixture
def tmp_flow_cells_demux_all_directory(tmp_path: Path, flow_cells_demux_all_dir: Path) -> Path:
    """
    Return the path to a temporary flow cells directory with flow cells ready for demultiplexing.
    Generates a copy of the original flow cells directory.
    This fixture is used for testing of the cg demutliplex all cmd.
    """
    original_dir = flow_cells_demux_all_dir
    tmp_dir = Path(tmp_path, "flow_cells_demux_all")

    return Path(shutil.copytree(original_dir, tmp_dir))


@pytest.fixture(name="tmp_flow_cell_directory_bcl2fastq")
def flow_cell_working_directory_bcl2fastq(
    bcl2fastq_flow_cell_dir: Path, tmp_flow_cells_directory: Path
) -> Path:
    """Return the path to a working directory that will be deleted after test is run.

    This is a path to a flow cell directory with the run parameters present.
    """
    return Path(tmp_flow_cells_directory, bcl2fastq_flow_cell_dir.name)


@pytest.fixture(name="tmp_flow_cell_directory_bclconvert")
def flow_cell_working_directory_bclconvert(
    bcl_convert_flow_cell_dir: Path, tmp_flow_cells_directory: Path
) -> Path:
    """Return the path to a working directory that will be deleted after test is run.
    This is a path to a flow cell directory with the run parameters present.
    """
    return Path(tmp_flow_cells_directory, bcl_convert_flow_cell_dir.name)


@pytest.fixture
def tmp_flow_cell_without_run_parameters_path(
    tmp_broken_flow_cells_directory: Path, hiseq_2500_custom_index_flow_cell_name: str
) -> Path:
    """This is a path to a flow cell directory with the run parameters missing."""
    return Path(tmp_broken_flow_cells_directory, hiseq_2500_custom_index_flow_cell_name)


@pytest.fixture
def tmp_novaseq_6000_pre_1_5_kits_flow_cell_without_sample_sheet_path(
    tmp_broken_flow_cells_directory: Path,
) -> Path:
    """This is a path to a flow cell directory with the sample sheet missing."""
    return Path(tmp_broken_flow_cells_directory, "190927_A00689_0069_BHLYWYDSXX")


@pytest.fixture
def tmp_novaseq_6000_post_1_5_kits_flow_cell_without_sample_sheet_path(
    tmp_broken_flow_cells_directory: Path,
) -> Path:
    """This is a path to a flow cell directory with the sample sheet missing."""
    return Path(tmp_broken_flow_cells_directory, "230912_A00187_1009_AHK33MDRX3")


@pytest.fixture
def tmp_novaseq_x_without_sample_sheet_flow_cell_path(
    tmp_broken_flow_cells_directory: Path,
) -> Path:
    """This is a path to a flow cell directory with the sample sheet missing."""
    return Path(tmp_broken_flow_cells_directory, "20231108_LH00188_0028_B22F52TLT3")


@pytest.fixture
def tmp_flow_cells_directory_malformed_sample_sheet(
    tmp_flow_cell_name_malformed_sample_sheet: str, tmp_flow_cells_directory: Path
) -> Path:
    """This is a path to a flow cell directory with a sample sheet with malformed headers."""
    return Path(tmp_flow_cells_directory, tmp_flow_cell_name_malformed_sample_sheet)


@pytest.fixture
def tmp_flow_cells_directory_ready_for_demultiplexing_bcl_convert(
    bcl_convert_flow_cell_full_name: str, tmp_flow_cells_directory: Path
) -> Path:
    """This is a path to a flow cell directory with the run parameters missing."""
    return Path(tmp_flow_cells_directory, bcl_convert_flow_cell_full_name)


@pytest.fixture
def tmp_flow_cells_directory_ready_for_demultiplexing_bcl2fastq(
    tmp_flow_cell_name_ready_for_demultiplexing_bcl2fastq: str, tmp_flow_cells_directory: Path
) -> Path:
    """This is a path to a flow cell directory with the run parameters missing."""
    return Path(tmp_flow_cells_directory, tmp_flow_cell_name_ready_for_demultiplexing_bcl2fastq)


# Temporary demultiplexed runs fixtures
@pytest.fixture(name="tmp_demultiplexed_runs_directory")
def tmp_demultiplexed_flow_cells_directory(tmp_path: Path, demultiplexed_runs: Path) -> Path:
    """Return the path to a temporary demultiplex-runs directory.
    Generates a copy of the original demultiplexed-runs
    """
    original_dir = demultiplexed_runs
    tmp_dir = Path(tmp_path, "demultiplexed-runs")
    return Path(shutil.copytree(original_dir, tmp_dir))


@pytest.fixture(name="tmp_demultiplexed_runs_bcl2fastq_directory")
def tmp_demultiplexed_runs_bcl2fastq_directory(
    tmp_demultiplexed_runs_directory: Path, bcl2fastq_flow_cell_dir: Path
) -> Path:
    """Return the path to a temporary demultiplex-runs bcl2fastq flow cell directory."""
    return Path(tmp_demultiplexed_runs_directory, bcl2fastq_flow_cell_dir.name)


@pytest.fixture(name="tmp_demultiplexed_runs_not_finished_directory")
def tmp_demultiplexed_runs_not_finished_flow_cells_directory(
    tmp_path: Path, demux_results_not_finished_dir: Path
) -> Path:
    """
    Return a temporary demultiplex-runs-unfinished path with an unfinished flow cell directory.
    Generates a copy of the original demultiplexed-runs-unfinished directory.
    """
    original_dir = demux_results_not_finished_dir
    tmp_dir = Path(tmp_path, "demultiplexed-runs-unfinished")
    return Path(shutil.copytree(original_dir, tmp_dir))


@pytest.fixture(name="demultiplexed_runs_unfinished_bcl2fastq_flow_cell_directory")
def demultiplexed_runs_bcl2fastq_flow_cell_directory(
    tmp_demultiplexed_runs_not_finished_directory: Path,
    bcl2fastq_flow_cell_full_name: str,
) -> Path:
    """Copy the content of a demultiplexed but not finished directory to a temporary location."""
    return Path(tmp_demultiplexed_runs_not_finished_directory, bcl2fastq_flow_cell_full_name)


@pytest.fixture(name="novaseq6000_bcl_convert_sample_sheet_path")
def novaseq6000_sample_sheet_path() -> Path:
    """Return the path to a NovaSeq 6000 BCL convert sample sheet."""
    return Path(
        "tests",
        "fixtures",
        "apps",
        "sequencing_metrics_parser",
        "230622_A00621_0864_AHY7FFDRX2",
        "Unaligned",
        "Reports",
        "SampleSheet.csv",
    )


# Directory fixtures


@pytest.fixture(scope="session")
def demultiplex_fixtures(apps_dir: Path) -> Path:
    """Return the path to the demultiplex fixture directory."""
    return Path(apps_dir, "demultiplexing")


@pytest.fixture(scope="session")
def raw_lims_sample_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to the raw samples fixture directory."""
    return Path(demultiplex_fixtures, "raw_lims_samples")


@pytest.fixture(scope="session")
def run_parameters_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to the run parameters fixture directory."""
    return Path(demultiplex_fixtures, "run_parameters")


@pytest.fixture(scope="session")
def demultiplexed_runs(demultiplex_fixtures: Path) -> Path:
    """Return the path to the demultiplexed flow cells fixture directory."""
    return Path(demultiplex_fixtures, "demultiplexed-runs")


@pytest.fixture(scope="session")
def flow_cells_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to the sequenced flow cells fixture directory."""
    return Path(demultiplex_fixtures, DemultiplexingDirsAndFiles.FLOW_CELLS_DIRECTORY_NAME)


@pytest.fixture(scope="session")
def broken_flow_cells_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to the broken or incomplete flow cells fixture directory."""
    return Path(demultiplex_fixtures, "flow_cells_broken")


@pytest.fixture(scope="session")
def nanopore_flow_cells_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to the sequenced flow cells fixture directory."""
    return Path(demultiplex_fixtures, NanoporeDirsAndFiles.DATA_DIRECTORY)


@pytest.fixture(scope="session")
def flow_cells_demux_all_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to the sequenced flow cells fixture directory."""
    return Path(demultiplex_fixtures, "flow_cells_demux_all")


@pytest.fixture(scope="session")
def demux_results_not_finished_dir(demultiplex_fixtures: Path) -> Path:
    """Return the path to a dir with demultiplexing results where nothing has been cleaned."""
    return Path(demultiplex_fixtures, "demultiplexed-runs-unfinished")


###


@pytest.fixture
def novaseq_6000_post_1_5_kits_flow_cell_path(tmp_flow_cells_directory: Path) -> Path:
    return Path(tmp_flow_cells_directory, "230912_A00187_1009_AHK33MDRX3")


@pytest.fixture
def novaseq_6000_post_1_5_kits_correct_sample_sheet_path(
    novaseq_6000_post_1_5_kits_flow_cell_path: Path,
) -> Path:
    return Path(novaseq_6000_post_1_5_kits_flow_cell_path, CORRECT_SAMPLE_SHEET)


@pytest.fixture
def novaseq_6000_pre_1_5_kits_flow_cell_path(tmp_flow_cells_directory: Path) -> Path:
    return Path(tmp_flow_cells_directory, "190927_A00689_0069_BHLYWYDSXX")


@pytest.fixture
def novaseq_6000_pre_1_5_kits_correct_sample_sheet_path(
    novaseq_6000_pre_1_5_kits_flow_cell_path: Path,
) -> Path:
    return Path(novaseq_6000_pre_1_5_kits_flow_cell_path, CORRECT_SAMPLE_SHEET)


@pytest.fixture
def novaseq_x_flow_cell_directory(tmp_flow_cells_directory: Path) -> Path:
    return Path(tmp_flow_cells_directory, "20231108_LH00188_0028_B22F52TLT3")


@pytest.fixture
def novaseq_x_correct_sample_sheet(novaseq_x_flow_cell_directory: Path) -> Path:
    return Path(novaseq_x_flow_cell_directory, CORRECT_SAMPLE_SHEET)


@pytest.fixture
def novaseq_x_manifest_file(novaseq_x_flow_cell_dir: Path) -> Path:
    """Return the path to a NovaSeqX manifest file."""
    return Path(novaseq_x_flow_cell_dir, "Manifest.tsv")


@pytest.fixture(scope="session")
def hiseq_x_single_index_flow_cell_dir(
    flow_cells_dir: Path, hiseq_x_single_index_flow_cell_name: str
) -> Path:
    """Return the path to a HiSeqX flow cell."""
    return Path(flow_cells_dir, hiseq_x_single_index_flow_cell_name)


@pytest.fixture(scope="session")
def hiseq_x_dual_index_flow_cell_dir(
    flow_cells_dir: Path, hiseq_x_dual_index_flow_cell_name: str
) -> Path:
    """Return the path to a HiSeqX flow cell."""
    return Path(flow_cells_dir, hiseq_x_dual_index_flow_cell_name)


@pytest.fixture(scope="session")
def hiseq_2500_dual_index_flow_cell_dir(
    flow_cells_dir: Path, hiseq_2500_dual_index_flow_cell_name: str
) -> Path:
    """Return the path to a HiSeq2500 flow cell."""
    return Path(flow_cells_dir, hiseq_2500_dual_index_flow_cell_name)


@pytest.fixture(scope="session")
def hiseq_2500_custom_index_flow_cell_dir(
    flow_cells_dir: Path, hiseq_2500_custom_index_flow_cell_name: str
) -> Path:
    """Return the path to a HiSeq2500 flow cell."""
    return Path(flow_cells_dir, hiseq_2500_custom_index_flow_cell_name)


@pytest.fixture
def novaseq_x_flow_cell_dir(flow_cells_dir: Path) -> Path:
    """Return the path to a NovaSeqX flow cell."""
    return Path(flow_cells_dir, "20231108_LH00188_0028_B22F52TLT3")


@pytest.fixture(scope="session")
def bcl2fastq_flow_cell_dir(flow_cells_dir: Path, bcl2fastq_flow_cell_full_name: str) -> Path:
    """Return the path to the bcl2fastq flow cell demultiplex fixture directory."""
    return Path(flow_cells_dir, bcl2fastq_flow_cell_full_name)


@pytest.fixture(scope="session")
def bcl_convert_flow_cell_dir(flow_cells_dir: Path, bcl_convert_flow_cell_full_name: str) -> Path:
    """Return the path to the bcl_convert flow cell demultiplex fixture directory."""
    return Path(flow_cells_dir, bcl_convert_flow_cell_full_name)


@pytest.fixture(scope="session")
def novaseq_bcl2fastq_sample_sheet_path(bcl2fastq_flow_cell_dir: Path) -> Path:
    """Return the path to a NovaSeq6000 Bcl2fastq sample sheet."""
    return Path(bcl2fastq_flow_cell_dir, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME)


@pytest.fixture(scope="session")
def novaseq_bcl_convert_sample_sheet_path(bcl_convert_flow_cell_dir: Path) -> Path:
    """Return the path to a NovaSeq6000 bcl_convert sample sheet."""
    return Path(bcl_convert_flow_cell_dir, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME)


@pytest.fixture(scope="session")
def run_parameters_wrong_instrument(run_parameters_dir: Path) -> Path:
    """Return a NovaSeqX run parameters file path with a wrong instrument value."""
    return Path(run_parameters_dir, "RunParameters_novaseq_X_wrong_instrument.xml")


@pytest.fixture(scope="session")
def hiseq_x_single_index_run_parameters_path(
    hiseq_x_single_index_flow_cell_dir: Path,
) -> Path:
    """Return the path to a HiSeqX run parameters file with single index."""
    return Path(
        hiseq_x_single_index_flow_cell_dir, DemultiplexingDirsAndFiles.RUN_PARAMETERS_CAMEL_CASE
    )


@pytest.fixture(scope="session")
def hiseq_x_dual_index_run_parameters_path(
    hiseq_x_dual_index_flow_cell_dir: Path,
) -> Path:
    """Return the path to a HiSeqX run parameters file with dual index."""
    return Path(
        hiseq_x_dual_index_flow_cell_dir, DemultiplexingDirsAndFiles.RUN_PARAMETERS_CAMEL_CASE
    )


@pytest.fixture(scope="session")
def hiseq_2500_dual_index_run_parameters_path(
    hiseq_2500_dual_index_flow_cell_dir: Path,
) -> Path:
    """Return the path to a HiSeq2500 run parameters file with dual index."""
    return Path(
        hiseq_2500_dual_index_flow_cell_dir, DemultiplexingDirsAndFiles.RUN_PARAMETERS_CAMEL_CASE
    )


@pytest.fixture(scope="session")
def hiseq_2500_custom_index_run_parameters_path(
    hiseq_2500_custom_index_flow_cell_dir: Path,
) -> Path:
    """Return the path to a HiSeq2500 run parameters file with custom index."""
    return Path(
        hiseq_2500_custom_index_flow_cell_dir, DemultiplexingDirsAndFiles.RUN_PARAMETERS_CAMEL_CASE
    )


@pytest.fixture(scope="session")
def novaseq_6000_run_parameters_path(bcl2fastq_flow_cell_dir: Path) -> Path:
    """Return the path to a NovaSeq6000 run parameters file."""
    return Path(bcl2fastq_flow_cell_dir, DemultiplexingDirsAndFiles.RUN_PARAMETERS_PASCAL_CASE)


@pytest.fixture
def novaseq_6000_run_parameters_pre_1_5_kits_path(
    novaseq_6000_pre_1_5_kits_flow_cell_path: Path,
) -> Path:
    """Return the path to a NovaSeq6000 pre 1.5 kit run parameters file."""
    return Path(
        novaseq_6000_pre_1_5_kits_flow_cell_path,
        DemultiplexingDirsAndFiles.RUN_PARAMETERS_PASCAL_CASE,
    )


@pytest.fixture
def novaseq_6000_run_parameters_post_1_5_kits_path(
    novaseq_6000_post_1_5_kits_flow_cell_path: Path,
) -> Path:
    """Return the path to a NovaSeq6000 post 1.5 kit run parameters file."""
    return Path(
        novaseq_6000_post_1_5_kits_flow_cell_path,
        DemultiplexingDirsAndFiles.RUN_PARAMETERS_PASCAL_CASE,
    )


@pytest.fixture
def novaseq_x_run_parameters_path(novaseq_x_flow_cell_dir: Path) -> Path:
    """Return the path to a NovaSeqX run parameters file."""
    return Path(novaseq_x_flow_cell_dir, DemultiplexingDirsAndFiles.RUN_PARAMETERS_PASCAL_CASE)


@pytest.fixture(scope="module")
def run_parameters_missing_versions_path(
    run_parameters_dir: Path,
) -> Path:
    """Return a NovaSeq6000 run parameters path without software and reagent kit versions."""
    return Path(run_parameters_dir, "RunParameters_novaseq_no_software_nor_reagent_version.xml")


@pytest.fixture(name="lims_novaseq_samples_file")
def lims_novaseq_samples_file(raw_lims_sample_dir: Path) -> Path:
    """Return the path to a file with sample info in lims format."""
    return Path(raw_lims_sample_dir, "raw_samplesheet_novaseq.json")


@pytest.fixture
def lims_novaseq_6000_samples_file(bcl2fastq_flow_cell_dir: Path) -> Path:
    """Return the path to the file with the raw samples of HVKJCDRXX flow cell in lims format."""
    return Path(bcl2fastq_flow_cell_dir, "HVKJCDRXX_raw.json")


@pytest.fixture(name="demultiplexed_flow_cell")
def demultiplexed_flow_cell(demultiplexed_runs: Path, bcl2fastq_flow_cell_full_name: str) -> Path:
    """Return the path to a demultiplexed flow cell with bcl2fastq."""
    return Path(demultiplexed_runs, bcl2fastq_flow_cell_full_name)


@pytest.fixture(name="bcl_convert_demultiplexed_flow_cell")
def bcl_convert_demultiplexed_flow_cell(
    demultiplexed_runs: Path, bcl_convert_flow_cell_full_name: str
) -> Path:
    """Return the path to a demultiplexed flow cell with BCLConvert."""
    return Path(demultiplexed_runs, bcl_convert_flow_cell_full_name)


# Fixtures for test demultiplex flow cell
@pytest.fixture
def tmp_empty_demultiplexed_runs_directory(tmp_demultiplexed_runs_directory) -> Path:
    return Path(tmp_demultiplexed_runs_directory, "empty")


@pytest.fixture(name="novaseqx_demultiplexed_flow_cell")
def novaseqx_demultiplexed_flow_cell(demultiplexed_runs: Path, novaseq_x_flow_cell_full_name: str):
    """Return the path to a demultiplexed NovaSeqX flow cell."""
    return Path(demultiplexed_runs, novaseq_x_flow_cell_full_name)



import logging

from pydantic import BaseModel, BeforeValidator, model_validator
from typing_extensions import Annotated

from cg.constants import NA_FIELD, REPORT_SUPPORTED_WORKFLOW
from cg.models.report.sample import ApplicationModel, SampleModel
from cg.models.report.validators import (
    get_analysis_type_as_string,
    get_date_as_string,
    get_list_as_string,
    get_path_as_string,
    get_report_string,
)

LOG = logging.getLogger(__name__)


class CustomerModel(BaseModel):
    """
    Customer model associated to the delivery report generated

    Attributes:
        name: customer name; source: statusDB/family/customer/name
        id: customer internal ID; source: statusDB/family/customer/internal_id
        invoice_address: customers invoice address; source: statusDB/family/customer/invoice_address
        scout_access: whether the customer has access to scout or not; source: statusDB/family/customer/scout_access
    """

    name: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    id: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    invoice_address: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    scout_access: bool | None = None


class ScoutReportFiles(BaseModel):
    """
    Model that describes the files uploaded to Scout and delivered to the customer

    Attributes:
        snv_vcf: SNV VCF file uploaded to Scout; source: HK
        snv_research_vcf: SNV research VCF file uploaded to Scout; source: HK
        sv_vcf: SV VCF file uploaded to Scout; source: HK
        sv_research_vcf: SV research VCF file uploaded to Scout; source: HK
        vcf_str: Short Tandem Repeat variants file (MIP-DNA specific); source: HK
        smn_tsv: SMN gene variants file (MIP-DNA specific); source: HK
        vcf_fusion: Converted RNA fusion file to SV VCF (RNAfusion specific); source: HK
    """

    snv_vcf: Annotated[str, BeforeValidator(get_path_as_string)] = NA_FIELD
    snv_research_vcf: Annotated[str, BeforeValidator(get_path_as_string)] = NA_FIELD
    sv_vcf: Annotated[str, BeforeValidator(get_path_as_string)] = NA_FIELD
    sv_research_vcf: Annotated[str, BeforeValidator(get_path_as_string)] = NA_FIELD
    vcf_str: Annotated[str, BeforeValidator(get_path_as_string)] = NA_FIELD
    smn_tsv: Annotated[str, BeforeValidator(get_path_as_string)] = NA_FIELD
    vcf_fusion: Annotated[str, BeforeValidator(get_path_as_string)] = NA_FIELD


class DataAnalysisModel(BaseModel):
    """
    Model that describes the workflow attributes used for the data analysis

    Attributes:
        customer_workflow: data analysis requested by the customer; source: StatusDB/family/data_analysis
        data_delivery: data delivery requested by the customer; source: StatusDB/family/data_delivery
        workflow: actual workflow used for analysis; source: statusDB/analysis/pipeline
        workflow_version: workflow version; source: statusDB/analysis/pipeline_version
        type: analysis type carried out; source: pipeline workflow
        genome_build: build version of the genome reference; source: pipeline workflow
        variant_callers: variant-calling filters; source: pipeline workflow
        panels: list of case specific panels; source: StatusDB/family/panels
        scout_files: list of file names uploaded to Scout
    """

    customer_workflow: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    data_delivery: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    workflow: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    workflow_version: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    type: Annotated[str, BeforeValidator(get_analysis_type_as_string)] = NA_FIELD
    genome_build: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    variant_callers: Annotated[str, BeforeValidator(get_list_as_string)] = NA_FIELD
    panels: Annotated[str, BeforeValidator(get_list_as_string)] = NA_FIELD
    scout_files: ScoutReportFiles

    @model_validator(mode="after")
    def check_supported_workflow(self) -> "DataAnalysisModel":
        """Check if the report generation supports a specific workflow and analysis type."""
        if self.workflow != self.customer_workflow:
            LOG.error(
                f"The analysis requested by the customer ({self.customer_workflow}) does not match the one "
                f"executed ({self.workflow})"
            )
            raise ValueError
        if self.workflow not in REPORT_SUPPORTED_WORKFLOW:
            LOG.error(f"The workflow {self.workflow} does not support delivery report generation")
            raise ValueError
        return self


class CaseModel(BaseModel):
    """
    Defines the case/family model

    Attributes:
        name: case name; source: StatusDB/family/name
        id: case ID; source: StatusDB/family/internal_id
        samples: list of samples associated to a case/family
        data_analysis: pipeline attributes
        applications: case associated unique applications
    """

    name: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    id: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    samples: list[SampleModel]
    data_analysis: DataAnalysisModel
    applications: list[ApplicationModel]


class ReportModel(BaseModel):
    """
    Model that defines the report attributes to render

    Attributes:
        customer: customer attributes
        version: delivery report version; source: StatusDB/analysis/family/analyses(/index)
        date: report generation date; source: CG runtime
        case: case attributes
        accredited: whether the report is accredited or not; source: all(StatusDB/application/accredited)
    """

    customer: CustomerModel
    version: Annotated[str, BeforeValidator(get_report_string)] = NA_FIELD
    date: Annotated[str, BeforeValidator(get_date_as_string)] = NA_FIELD
    case: CaseModel
    accredited: bool | None = None



import logging
from pathlib import Path
from typing import Type

from pydantic import TypeAdapter

from cg.apps.demultiplex.sample_sheet.sample_models import (
    FlowCellSample,
    FlowCellSampleBcl2Fastq,
    FlowCellSampleBCLConvert,
)
from cg.apps.demultiplex.sample_sheet.sample_sheet_models import SampleSheet
from cg.constants.constants import FileFormat
from cg.constants.demultiplexing import SampleSheetBcl2FastqSections, SampleSheetBCLConvertSections
from cg.exc import SampleSheetError
from cg.io.controller import ReadFile

LOG = logging.getLogger(__name__)


def validate_samples_are_unique(samples: list[FlowCellSample]) -> None:
    """Validate that each sample only exists once."""
    sample_ids: set = set()
    for sample in samples:
        sample_id: str = sample.sample_id.split("_")[0]
        if sample_id in sample_ids:
            message: str = f"Sample {sample.sample_id} exists multiple times in sample sheet"
            LOG.warning(message)
            raise SampleSheetError(message)
        sample_ids.add(sample_id)


def validate_samples_unique_per_lane(samples: list[FlowCellSample]) -> None:
    """Validate that each sample only exists once per lane in a sample sheet."""
    sample_by_lane: dict[int, list[FlowCellSample]] = get_samples_by_lane(samples)
    for lane, lane_samples in sample_by_lane.items():
        LOG.debug(f"Validate that samples are unique in lane: {lane}")
        validate_samples_are_unique(samples=lane_samples)


def get_sample_sheet_from_file(infile: Path) -> SampleSheet:
    """Parse and validate a sample sheet from file."""
    sample_sheet_content: list[list[str]] = ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=infile
    )
    sample_type: Type[FlowCellSample] = get_sample_type(infile)

    return get_validated_sample_sheet(
        sample_sheet_content=sample_sheet_content,
        sample_type=sample_type,
    )


def get_sample_type(sample_sheet_path: Path) -> Type[FlowCellSample]:
    """Returns the sample type based on the header of the given sample sheet."""
    sample_sheet_content: list[list[str]] = ReadFile.get_content_from_file(
        file_format=FileFormat.CSV, file_path=sample_sheet_path
    )
    for row in sample_sheet_content:
        if not row:
            continue
        if SampleSheetBCLConvertSections.Data.HEADER in row[0]:
            LOG.info("Sample sheet was generated for BCL Convert")
            return FlowCellSampleBCLConvert
        if SampleSheetBcl2FastqSections.Data.HEADER in row[0]:
            LOG.info("Sample sheet was generated for BCL2FASTQ")
            return FlowCellSampleBcl2Fastq
    raise SampleSheetError("Could not determine sample sheet type")


def get_validated_sample_sheet(
    sample_sheet_content: list[list[str]],
    sample_type: Type[FlowCellSample],
) -> SampleSheet:
    """Return a validated sample sheet object."""
    raw_samples: list[dict[str, str]] = get_raw_samples(sample_sheet_content=sample_sheet_content)
    adapter = TypeAdapter(list[sample_type])
    samples = adapter.validate_python(raw_samples)
    validate_samples_unique_per_lane(samples=samples)
    return SampleSheet(samples=samples)


def get_raw_samples(sample_sheet_content: list[list[str]]) -> list[dict[str, str]]:
    """Return the samples in a sample sheet as a list of dictionaries."""
    header: list[str] = []
    raw_samples: list[dict[str, str]] = []

    for line in sample_sheet_content:
        # Skip lines that are too short to contain samples
        if len(line) <= 5:
            continue
        if line[0] in [
            SampleSheetBcl2FastqSections.Data.FLOW_CELL_ID.value,
            SampleSheetBCLConvertSections.Data.LANE.value,
        ]:
            header = line
            continue
        if not header:
            continue
        raw_samples.append(dict(zip(header, line)))
    if not header:
        message = "Could not find header in sample sheet"
        LOG.warning(message)
        raise SampleSheetError(message)
    if not raw_samples:
        message = "Could not find any samples in sample sheet"
        LOG.warning(message)
        raise SampleSheetError(message)
    return raw_samples


def get_samples_by_lane(
    samples: list[FlowCellSample],
) -> dict[int, list[FlowCellSample]]:
    """Group and return samples by lane."""
    LOG.debug("Order samples by lane")
    sample_by_lane: dict[int, list[FlowCellSample]] = {}
    for sample in samples:
        if sample.lane not in sample_by_lane:
            sample_by_lane[sample.lane] = []
        sample_by_lane[sample.lane].append(sample)
    return sample_by_lane



"""Common CLI file generation functions"""

import click

from cg.cli.generate.report.base import (
    generate_available_delivery_reports,
    generate_delivery_report,
)


@click.group()
def generate():
    """Generates and/or modifies files."""


generate.add_command(generate_delivery_report)
generate.add_command(generate_available_delivery_reports)



from http import HTTPStatus
from flask.testing import FlaskClient

from cg.store.models import Case


def test_get_delivery_message(client: FlaskClient, case: Case):
    # GIVEN a case

    # WHEN a request is made to get a delivery message for the case
    response = client.get(f"/api/v1/cases/{case.internal_id}/delivery_message")

    # THEN the response should be successful
    assert response.status_code == HTTPStatus.OK

    # THEN the response contains a message
    assert response.json["message"]



import pytest


@pytest.fixture
def tmp_flow_cell_name_malformed_sample_sheet() -> str:
    """ "Returns the name of a flow cell directory ready for demultiplexing with BCL convert.
    Contains a sample sheet with malformed headers.
    """
    return "201203_A00689_0200_AHVKJCDRXY"


@pytest.fixture(name="tmp_flow_cell_name_ready_for_demultiplexing_bcl2fastq")
def tmp_flow_cell_name_ready_for_demultiplexing_bcl2fastq() -> str:
    """Returns the name of a flow cell directory ready for demultiplexing with bcl2fastq."""
    return "211101_D00483_0615_AHLG5GDRXY"


@pytest.fixture(scope="session")
def flow_cell_name_demultiplexed_with_bcl2fastq() -> str:
    """Return the name of a flow cell that has been demultiplexed with BCL2Fastq."""
    return "HHKVCALXX"


@pytest.fixture(scope="session")
def flow_cell_directory_name_demultiplexed_with_bcl2fastq(
    flow_cell_name_demultiplexed_with_bcl2fastq: str,
) -> str:
    """Return the name of a flow cell directory that has been demultiplexed with BCL2Fastq."""
    return f"170407_ST-E00198_0209_B{flow_cell_name_demultiplexed_with_bcl2fastq}"


@pytest.fixture(scope="session")
def flow_cell_name_demultiplexed_with_bcl_convert() -> str:
    return "HY7FFDRX2"


@pytest.fixture(scope="session")
def flow_cell_directory_name_demultiplexed_with_bcl_convert(
    flow_cell_name_demultiplexed_with_bcl_convert: str,
) -> str:
    return f"230504_A00689_0804_B{flow_cell_name_demultiplexed_with_bcl_convert}"


@pytest.fixture(scope="session")
def hiseq_x_single_index_flow_cell_name() -> str:
    """Return the full name of a HiSeqX flow cell with only one index."""
    return "170517_ST-E00266_0210_BHJCFFALXX"


@pytest.fixture(scope="session")
def hiseq_x_dual_index_flow_cell_name() -> str:
    """Return the full name of a HiSeqX flow cell with two indexes."""
    return "180508_ST-E00269_0269_AHL32LCCXY"


@pytest.fixture(scope="session")
def hiseq_2500_dual_index_flow_cell_name() -> str:
    """Return the full name of a HiSeq2500 flow cell with double indexes."""
    return "181005_D00410_0735_BHM2LNBCX2"


@pytest.fixture(scope="session")
def hiseq_2500_custom_index_flow_cell_name() -> str:
    """Return the full name of a HiSeq2500 flow cell with double indexes."""
    return "180509_D00450_0598_BHGYFNBCX2"


@pytest.fixture(scope="session")
def bcl2fastq_flow_cell_full_name() -> str:
    """Return full flow cell name."""
    return "201203_D00483_0200_AHVKJCDRXX"


@pytest.fixture(scope="session")
def bcl_convert_flow_cell_full_name() -> str:
    """Return the full name of a bcl_convert flow cell."""
    return "211101_A00187_0615_AHLG5GDRZZ"


@pytest.fixture(scope="session")
def novaseq_x_flow_cell_full_name() -> str:
    """Return the full name of a NovaSeqX flow cell."""
    return "20230508_LH00188_0003_A22522YLT3"


# Lists


@pytest.fixture(scope="session")
def bcl_convert_demultiplexed_flow_cell_sample_internal_ids() -> list[str]:
    """
    Sample id:s present in sample sheet for dummy flow cell demultiplexed with BCL Convert in
    cg/tests/fixtures/apps/demultiplexing/demultiplexed-runs/230504_A00689_0804_BHY7FFDRX2.
    """
    return ["ACC11927A2", "ACC11927A5"]


@pytest.fixture(scope="session")
def bcl2fastq_demultiplexed_flow_cell_sample_internal_ids() -> list[str]:
    """
    Sample id:s present in sample sheet for dummy flow cell demultiplexed with BCL Convert in
    cg/tests/fixtures/apps/demultiplexing/demultiplexed-runs/170407_A00689_0209_BHHKVCALXX.
    """
    return ["SVE2528A1"]



from pydantic import BaseModel

from cg.constants import Workflow


class Order(BaseModel):
    customer_id: str
    ticket_id: int
    order_date: str
    order_id: int
    workflow: Workflow


class OrdersResponse(BaseModel):
    orders: list[Order]



"""Module for deliver and rsync customer inbox on the HPC to customer inbox on the delivery
server """

import datetime
import logging
import os
import re
import shutil
from pathlib import Path

from cg.constants.delivery import INBOX_NAME
from cg.exc import CgError
from cg.meta.meta import MetaAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Case, Sample

LOG = logging.getLogger(__name__)
PREFIX_TO_CONCATENATE = ["MWG", "MWL", "MWM", "MWR", "MWX"]


class DeliverTicketAPI(MetaAPI):
    def __init__(self, config: CGConfig):
        super().__init__(config)
        self.delivery_path: Path = Path(config.delivery_path)

    def get_all_cases_from_ticket(self, ticket: str) -> list[Case]:
        return self.status_db.get_cases_by_ticket_id(ticket_id=ticket)

    def get_inbox_path(self, ticket: str) -> Path:
        cases: list[Case] = self.get_all_cases_from_ticket(ticket=ticket)
        if not cases:
            raise CgError(
                f"The customer id was not identified since no cases for ticket {ticket} was found"
            )
        customer_id: str = cases[0].customer.internal_id
        return Path(self.delivery_path, customer_id, INBOX_NAME, ticket)

    def check_if_upload_is_needed(self, ticket: str) -> bool:
        customer_inbox: Path = self.get_inbox_path(ticket=ticket)
        LOG.info(f"Checking if path exist: {customer_inbox}")
        if customer_inbox.exists():
            LOG.info(f"Could find path: {customer_inbox}")
            return False
        LOG.info(f"Could not find path: {customer_inbox}")
        return True

    def generate_date_tag(self, ticket: str) -> datetime.datetime:
        cases: list[Case] = self.get_all_cases_from_ticket(ticket=ticket)
        return cases[0].ordered_at

    def generate_output_filename(
        self, date: datetime.datetime, dir_path: Path, read_direction: int
    ) -> Path:
        base_name = Path("_".join([dir_path.name, str(read_direction)]))
        if date:
            base_name = Path("_".join([str(date.strftime("%y%m%d")), str(base_name)]))
        fastq_file_name = base_name.with_suffix(".fastq.gz")
        return Path(dir_path, fastq_file_name)

    @staticmethod
    def sort_files(files: list[Path]) -> list[Path]:
        files_map = {file_path.name: file_path for file_path in files}
        sorted_names = sorted(list(files_map.keys()))
        return [files_map[file_name] for file_name in sorted_names]

    def get_current_read_direction(self, dir_path: Path, read_direction: int) -> list[Path]:
        same_direction = []
        direction_string = ".+_R" + str(read_direction) + "_[0-9]+.fastq.gz"
        direction_pattern = re.compile(direction_string)
        file_path: Path
        for file_path in dir_path.iterdir():
            if direction_pattern.match(str(file_path)):
                same_direction.append(file_path)
        return self.sort_files(files=same_direction)

    def get_total_size(self, files: list[Path]) -> int:
        total_size = 0
        for file in files:
            total_size += file.stat().st_size
        return total_size

    def concatenate_same_read_direction(self, reads: list[Path], output: Path) -> None:
        with open(output, "wb") as write_file_obj:
            for file in reads:
                with open(file, "rb") as file_descriptor:
                    shutil.copyfileobj(file_descriptor, write_file_obj)

    def remove_files(self, reads: list[Path]) -> None:
        for file in reads:
            LOG.info(f"Removing file: {file}")
            file.unlink()

    def get_samples_from_ticket(self, ticket: str) -> list:
        all_samples = []
        cases: list[Case] = self.get_all_cases_from_ticket(ticket=ticket)
        for case in cases:
            for link_obj in case.links:
                all_samples.append(link_obj.sample.name)
        return all_samples

    def report_missing_samples(self, ticket: str, dry_run: bool) -> None:
        customer_inbox: Path = self.get_inbox_path(ticket=ticket)
        missing_samples = []
        all_samples: list = self.get_samples_from_ticket(ticket=ticket)
        if not customer_inbox.exists() and dry_run:
            LOG.info(f"Dry run, will not search for missing data in: {customer_inbox}")
            return
        if not customer_inbox.exists():
            LOG.info(
                f"The path {customer_inbox} do not exist, no search for missing data will be done"
            )
            return
        for dir_path in customer_inbox.iterdir():
            if len(os.listdir(dir_path)) == 0 and os.path.basename(dir_path) in all_samples:
                missing_samples.append(os.path.basename(dir_path))
        if len(missing_samples) > 0:
            LOG.info("No data delivered for sample(s):")
            for sample in missing_samples:
                LOG.info(sample)
        else:
            LOG.info("Data has been delivered for all samples")

    def concatenate(self, ticket: str, dry_run: bool) -> None:
        customer_inbox: Path = self.get_inbox_path(ticket=ticket)
        date: datetime.datetime = self.generate_date_tag(ticket=ticket)
        if not customer_inbox.exists() and dry_run:
            LOG.info(f"Dry run, nothing will be concatenated in: {customer_inbox}")
            return
        if not customer_inbox.exists():
            LOG.info(f"The path {customer_inbox} do not exist, nothing will be concatenated")
            return
        for dir_path in customer_inbox.iterdir():
            if len(os.listdir(dir_path)) == 0:
                LOG.info(f"Empty folder found: {dir_path}")
                continue
            if not dir_path.is_dir():
                continue
            for read_direction in [1, 2]:
                same_direction: list[Path] = self.get_current_read_direction(
                    dir_path=dir_path, read_direction=read_direction
                )
                total_size: int = self.get_total_size(files=same_direction)
                output: Path = self.generate_output_filename(
                    date=date, dir_path=dir_path, read_direction=read_direction
                )
                if dry_run:
                    for file in same_direction:
                        LOG.info(f"Dry run activated, {file} will not be appended to {output}")
                else:
                    LOG.info(f"Concatenating sample: {dir_path.name}")
                    self.concatenate_same_read_direction(reads=same_direction, output=output)
                if dry_run:
                    continue
                concatenated_size = output.stat().st_size
                if total_size != concatenated_size:
                    raise CgError("WARNING data lost in concatenation")

                LOG.info(
                    "QC PASSED: Total size for files used in concatenation match the size of the concatenated file"
                )
                self.remove_files(reads=same_direction)

    def get_app_tag(self, samples: list) -> str:
        app_tag = samples[0].application_version.application.tag
        return app_tag

    def check_if_concatenation_is_needed(self, ticket: str) -> bool:
        cases: list[Case] = self.get_all_cases_from_ticket(ticket=ticket)
        case_id = cases[0].internal_id
        case_obj = self.status_db.get_case_by_internal_id(internal_id=case_id)
        samples: list[Sample] = [link.sample for link in case_obj.links]
        app_tag = self.get_app_tag(samples=samples)
        for prefix in PREFIX_TO_CONCATENATE:
            if app_tag.startswith(prefix):
                LOG.info(
                    f"Identified {app_tag} as application tag, i.e. the fastqs should be concatenated",
                )
                return True
        LOG.info(
            f"The following application tag was identified: {app_tag}, concatenation will be skipped",
        )
        return False



from cg.services.delivery_message.messages.delivery_message import DeliveryMessage
from cg.services.delivery_message.messages.utils import get_fastq_delivery_path, get_scout_link
from cg.store.models import Case


class FastqScoutMessage(DeliveryMessage):
    def create_message(self, case: Case) -> str:
        scout_link: str = get_scout_link(case)
        delivery_path: str = get_fastq_delivery_path(case)
        return (
            f"Hello,\n\n "
            f"The results have been uploaded to Scout for the following case::\n\n"
            f"{scout_link}\n\n"
            f"The fastq files are currently being uploaded to your inbox on Caesar:\n\n"
            f"{delivery_path}"
        )



import logging
from pathlib import Path
from typing import Type

import pytest
from _pytest.fixtures import FixtureRequest
from pydantic import BaseModel

from cg.constants.demultiplexing import (
    NO_REVERSE_COMPLEMENTS_INDEX_SETTINGS,
    NOVASEQ_6000_POST_1_5_KITS_INDEX_SETTINGS,
    NOVASEQ_X_INDEX_SETTINGS,
    IndexSettings,
    RunParametersXMLNodes,
)
from cg.constants.sequencing import Sequencers
from cg.exc import RunParametersError, XMLError
from cg.models.demultiplex.run_parameters import (
    RunParameters,
    RunParametersHiSeq,
    RunParametersNovaSeq6000,
    RunParametersNovaSeqX,
)
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData


@pytest.mark.parametrize(
    "run_parameters_path_fixture",
    [
        "hiseq_x_single_index_run_parameters_path",
        "hiseq_2500_dual_index_run_parameters_path",
        "novaseq_6000_run_parameters_path",
        "novaseq_x_run_parameters_path",
    ],
)
def test_run_parameters_parent_class_fails(
    run_parameters_path_fixture: str, request: FixtureRequest
):
    """Test that trying to instantiate the parent RunParameters class raises an error."""
    # GIVEN a valid path for a run parameters file
    run_parameters_path: Path = request.getfixturevalue(run_parameters_path_fixture)

    # WHEN trying to instantiate the parent class with a RunParameters path
    with pytest.raises(NotImplementedError) as exc_info:
        # THEN an NotImplementedError is raised
        RunParameters(run_parameters_path=run_parameters_path)
    assert "Parent class instantiated" in str(exc_info)


@pytest.mark.parametrize(
    "run_parameters_path, constructor, sequencer",
    [
        ("hiseq_x_single_index_run_parameters_path", RunParametersHiSeq, Sequencers.HISEQX),
        ("hiseq_2500_dual_index_run_parameters_path", RunParametersHiSeq, Sequencers.HISEQGA),
        ("novaseq_6000_run_parameters_path", RunParametersNovaSeq6000, Sequencers.NOVASEQ),
        ("novaseq_x_run_parameters_path", RunParametersNovaSeqX, Sequencers.NOVASEQX),
    ],
)
def test_run_parameters_init(
    run_parameters_path: str,
    constructor: Type[RunParameters],
    sequencer: str,
    request: FixtureRequest,
):
    """Test that the RunParameters class is instantiated correctly."""
    # GIVEN a valid path for a run parameters file
    run_parameters_path: Path = request.getfixturevalue(run_parameters_path)

    # WHEN initialising the correct RunParameters class
    run_parameters: RunParameters = constructor(run_parameters_path=run_parameters_path)

    # THEN the object is instantiated correctly and is of the correct type
    assert isinstance(run_parameters, constructor)
    # THEN the sequencer is correct
    assert run_parameters.sequencer == sequencer


class RunParametersScenario(BaseModel):
    wrong_run_parameters_path_fixture: str
    constructor: Type[RunParameters]
    error_msg: str


@pytest.mark.parametrize(
    "scenario",
    [
        RunParametersScenario(
            wrong_run_parameters_path_fixture="novaseq_6000_run_parameters_path",
            constructor=RunParametersHiSeq,
            error_msg=f"Could not find node {RunParametersXMLNodes.APPLICATION_NAME} in the run parameters file.",
        ),
        RunParametersScenario(
            wrong_run_parameters_path_fixture="novaseq_x_run_parameters_path",
            constructor=RunParametersHiSeq,
            error_msg=f"Could not find node {RunParametersXMLNodes.APPLICATION_NAME} in the run parameters file.",
        ),
        RunParametersScenario(
            wrong_run_parameters_path_fixture="novaseq_x_run_parameters_path",
            constructor=RunParametersNovaSeq6000,
            error_msg=f"The file parsed does not correspond to {RunParametersXMLNodes.NOVASEQ_6000_APPLICATION}",
        ),
        RunParametersScenario(
            wrong_run_parameters_path_fixture="hiseq_x_single_index_run_parameters_path",
            constructor=RunParametersNovaSeq6000,
            error_msg=f"Could not find node {RunParametersXMLNodes.APPLICATION} in the run parameters file.",
        ),
        RunParametersScenario(
            wrong_run_parameters_path_fixture="hiseq_x_single_index_run_parameters_path",
            constructor=RunParametersNovaSeqX,
            error_msg=f"Could not find node {RunParametersXMLNodes.INSTRUMENT_TYPE} in the run parameters file.",
        ),
        RunParametersScenario(
            wrong_run_parameters_path_fixture="novaseq_6000_run_parameters_path",
            constructor=RunParametersNovaSeqX,
            error_msg=f"Could not find node {RunParametersXMLNodes.INSTRUMENT_TYPE} in the run parameters file.",
        ),
    ],
    ids=[
        "NovaSeq 6000 file, HiSeq constructor",
        "NovaSeq X file, HiSeq constructor",
        "NovaSeq X file, NovaSeq6000 constructor",
        "HiSeq X file, NovaSeq600 constructor",
        "HiSeq X file, NovaSeqX constructor",
        "NovaSeq6000 constructor, NovaSeqX constructor",
    ],
)
def test_run_parameters_wrong_file(scenario: RunParametersScenario, request: FixtureRequest):
    """Test that creating a RunParameters object with a wrong file fails."""
    # GIVEN a run parameters path from an instrument different from the expected
    wrong_run_parameters_path: Path = request.getfixturevalue(
        scenario.wrong_run_parameters_path_fixture
    )

    # WHEN trying to create a RunParameters object with the file
    with pytest.raises(RunParametersError) as exc_info:
        # THEN a RunParametersError is raised
        scenario.constructor(run_parameters_path=wrong_run_parameters_path)
    assert scenario.error_msg in str(exc_info.value)


@pytest.mark.parametrize(
    "run_parameters_fixture",
    [
        "hiseq_2500_dual_index_run_parameters",
        "hiseq_x_single_index_run_parameters",
        "novaseq_x_run_parameters",
    ],
    ids=["HiSeq 2500 dual index", "HiSeq X single index", "NovaSeq X"],
)
def test_reagent_kit_version_hiseq_and_novaseq_x(
    run_parameters_fixture: str, request: FixtureRequest
):
    """Test that getting reagent kit version from a HiSeq or NovaSeqX RunParameters returns None."""
    # GIVEN a valid RunParameters object
    run_parameters: RunParameters = request.getfixturevalue(run_parameters_fixture)

    # WHEN fetching the reagent kit version

    # THEN the reagent kit version is None
    assert not run_parameters.reagent_kit_version


def test_reagent_kit_version_novaseq_6000(novaseq_6000_run_parameters: RunParametersNovaSeq6000):
    """Test that getting reagent kit version from a correct file returns an expected value."""
    # GIVEN a valid RunParameters object for NovaSeq6000

    # WHEN fetching the reagent kit version
    reagent_kit_version: str = novaseq_6000_run_parameters.reagent_kit_version

    # THEN the reagent kit version exists and is not "unknown"
    assert reagent_kit_version
    assert reagent_kit_version != RunParametersXMLNodes.UNKNOWN_REAGENT_KIT_VERSION


@pytest.mark.parametrize(
    "run_parameters_fixture",
    [
        "hiseq_2500_dual_index_run_parameters",
        "hiseq_x_single_index_run_parameters",
        "novaseq_x_run_parameters",
    ],
)
def test_control_software_version_hiseq_and_novaseq_x(
    run_parameters_fixture: str, request: FixtureRequest
):
    """Test that getting control software version from HiSeq/NovaSeqX RunParameters returns None."""
    # GIVEN a valid RunParameters object
    run_parameters: RunParameters = request.getfixturevalue(run_parameters_fixture)

    # WHEN fetching the control software version

    # THEN the control software version is None
    assert not run_parameters.control_software_version


def test_control_software_version_novaseq_6000(
    novaseq_6000_run_parameters: RunParametersNovaSeq6000,
):
    """Test that getting control software version from a correct file returns an expected value."""
    # GIVEN a valid RunParameters object for NovaSeq6000

    # WHEN fetching the control software version
    control_software_version: str = novaseq_6000_run_parameters.control_software_version

    # THEN the control software version is a non-empty string
    assert isinstance(control_software_version, str)
    assert control_software_version != ""


def test_novaseq_6000_no_version(run_parameters_missing_versions_path: Path, caplog):
    """Test that creating a Run Parameters object from a file without version fails."""
    caplog.set_level(logging.INFO)
    # GIVEN the path of a RunParameters file without control software version

    # WHEN instantiating the object
    with pytest.raises(XMLError):
        # THEN assert that an exception was raised since the control software version was not found
        RunParametersNovaSeq6000(run_parameters_missing_versions_path)

    assert (
        f"Could not find node with name {RunParametersXMLNodes.APPLICATION_VERSION} in XML tree"
        in caplog.text
    )


@pytest.mark.parametrize(
    "run_parameters_fixture",
    [
        "hiseq_2500_dual_index_run_parameters",
        "hiseq_x_single_index_run_parameters",
        "novaseq_6000_run_parameters",
        "novaseq_x_run_parameters",
    ],
)
def test_get_cycles(run_parameters_fixture: str, request: FixtureRequest):
    """Test that the read and index cycles are read correctly for any RunParameters object."""
    # GIVEN a RunParameters object
    run_parameters: RunParameters = request.getfixturevalue(run_parameters_fixture)

    # WHEN getting any read cycle
    read_cycles: list[int] = [
        run_parameters.get_read_1_cycles(),
        run_parameters.get_read_2_cycles(),
        run_parameters.get_index_1_cycles(),
        run_parameters.get_index_2_cycles(),
    ]

    # THEN all read cycles are non-negative integers
    for cycles in read_cycles:
        assert isinstance(cycles, int)
        assert cycles >= 0


@pytest.mark.parametrize(
    "flow_cell_fixture, expected_result",
    [
        ("novaseq_6000_pre_1_5_kits_flow_cell", False),
        ("novaseq_6000_post_1_5_kits_flow_cell", True),
    ],
)
def test_is_novaseq6000_post_1_5_kit(
    flow_cell_fixture: str, expected_result: bool, request: FixtureRequest
):
    """Test that the correct index settings are returned for each NovaSeq flow cell type."""
    # GIVEN run parameters from a flow cell
    flow_cell: FlowCellDirectoryData = request.getfixturevalue(flow_cell_fixture)
    # WHEN checking if the flow cell was sequenced after the NovaSeq 6000 1.5 kits
    result: bool = flow_cell.run_parameters._is_novaseq6000_post_1_5_kit()
    # THEN the correct index settings are returned
    assert result == expected_result


@pytest.mark.parametrize(
    "flow_cell, correct_settings",
    [
        ("novaseq_6000_pre_1_5_kits_flow_cell", NO_REVERSE_COMPLEMENTS_INDEX_SETTINGS),
        ("novaseq_6000_post_1_5_kits_flow_cell", NOVASEQ_6000_POST_1_5_KITS_INDEX_SETTINGS),
        ("novaseq_x_flow_cell", NOVASEQ_X_INDEX_SETTINGS),
    ],
)
def test_get_index_settings(
    correct_settings: IndexSettings,
    flow_cell: str,
    request: FixtureRequest,
):
    """Test that the correct index settings are returned for each NovaSeq flow cell type."""
    # GIVEN run parameters for a flow cell
    flow_cell: FlowCellDirectoryData = request.getfixturevalue(flow_cell)
    # WHEN getting the index settings
    settings: IndexSettings = flow_cell.run_parameters.index_settings
    # THEN the correct index settings are returned
    assert settings == correct_settings



"""Tests for MicroSALT analysis."""

from datetime import datetime
from pathlib import Path

from mock import MagicMock

from cg.apps.lims.api import LimsAPI
from cg.meta.workflow.microsalt import MicrosaltAnalysisAPI
from cg.meta.workflow.microsalt.utils import get_project_directory_date
from cg.models.cg_config import CGConfig
from cg.store.models import Case


def test_get_cases_to_store_pass(
    qc_microsalt_context: CGConfig,
    mocker,
    microsalt_qc_pass_run_dir_path: Path,
    metrics_file_passing_qc: Path,
):
    """Test get cases to store for a microsalt case that passes QC."""

    # GIVEN a store with a QC ready microsalt case that will pass QC
    microsalt_api: MicrosaltAnalysisAPI = qc_microsalt_context.meta_apis["analysis_api"]
    mocker.patch.object(
        MicrosaltAnalysisAPI, "get_metrics_file_path", return_value=metrics_file_passing_qc
    )
    mocker.patch.object(
        MicrosaltAnalysisAPI, "get_case_path", return_value=microsalt_qc_pass_run_dir_path
    )

    # WHEN retrieving cases to store
    cases_to_store: list[Case] = microsalt_api.get_cases_to_store()

    # THEN cases should returned
    assert cases_to_store


def test_get_date_from_project_directory():
    # GIVEN a microsalt analysis run directory name
    run_dir_name = "ACC13796_2024.2.5_15.58.22"

    # WHEN parsing the project directory date
    run_date: datetime = get_project_directory_date(run_dir_name)

    # THEN the date should be parsed correctly
    assert run_date == datetime(2024, 2, 5, 15, 58, 22)



from cg.exc import OrderNotFoundError
from cg.models.orders.order import OrderIn
from cg.server.dto.orders.orders_request import OrdersRequest
from cg.server.dto.orders.orders_response import Order as OrderResponse
from cg.server.dto.orders.orders_response import OrdersResponse
from cg.services.orders.utils import create_order_response, create_orders_response
from cg.store.models import Order
from cg.store.store import Store


class OrderService:
    def __init__(self, store: Store) -> None:
        self.store = store

    def get_order(self, order_id: int) -> OrderResponse:
        order: Order | None = self.store.get_order_by_id(order_id)
        if not order:
            raise OrderNotFoundError(f"Order {order_id} not found.")
        return create_order_response(order)

    def get_orders(self, orders_request: OrdersRequest) -> OrdersResponse:
        orders: list[Order] = self._get_orders(orders_request)
        return create_orders_response(orders)

    def _get_orders(self, orders_request: OrdersRequest) -> list[Order]:
        """Returns a list of entries in the table Order."""
        return self.store.get_orders_by_workflow(
            workflow=orders_request.workflow, limit=orders_request.limit
        )

    def create_order(self, order_data: OrderIn) -> OrderResponse:
        order: Order = self.store.add_order(order_data)
        return create_order_response(order)



from cg.services.delivery_message.messages.delivery_message import DeliveryMessage
from cg.services.delivery_message.messages.utils import (
    get_fastq_delivery_path,
    get_pangolin_delivery_path,
)
from cg.store.models import Case


class CovidMessage(DeliveryMessage):
    def create_message(self, case: Case) -> str:
        delivery_path: str = get_fastq_delivery_path(case)
        pangolin_delivery_path: str = get_pangolin_delivery_path(case)
        return (
            f"Hello,\n\n "
            f"The analysis is now complete.\n\n"
            f"The result files are being uploaded on caesar at:\n\n"
            f"{delivery_path}\n\n"
            f"and the .csv files with pangolin-type have been sent to\n\n"
            f"{pangolin_delivery_path}"
        )



from enum import Enum
from typing import Callable

from sqlalchemy.orm import Query

from cg.store.models import Order


def filter_orders_by_workflow(orders: Query, workflow: str, **kwargs) -> Query:
    """Return orders filtered on workflow."""
    return orders.filter(Order.workflow == workflow) if workflow else orders


def filter_orders_by_id(orders: Query, id: int, **kwargs) -> Query:
    """Return orders filtered on id."""
    return orders.filter(Order.id == id)


def apply_order_filters(
    filter_functions: list[Callable], orders: Query, id: int = None, workflow: str = None
) -> Query:
    """Apply filtering functions to the order queries and return filtered results."""
    for filter_function in filter_functions:
        orders: Query = filter_function(orders=orders, id=id, workflow=workflow)
    return orders


class OrderFilter(Enum):
    """Define order filter functions."""

    FILTER_ORDERS_BY_ID: Callable = filter_orders_by_id
    FILTER_ORDERS_BY_WORKFLOW: Callable = filter_orders_by_workflow



"""Tests cli methods to create the case config for Taxprofiler."""

import logging
from pathlib import Path

from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from cg.cli.workflow.taxprofiler.base import config_case
from cg.constants import EXIT_SUCCESS
from cg.constants.constants import FileFormat
from cg.io.controller import ReadFile
from cg.models.cg_config import CGConfig
from cg.models.taxprofiler.taxprofiler import (
    TaxprofilerParameters,
    TaxprofilerSampleSheetEntry,
)


def test_config_case_default_parameters(
    cli_runner: CliRunner,
    taxprofiler_context: CGConfig,
    taxprofiler_case_id: str,
    taxprofiler_sample_sheet_path: Path,
    taxprofiler_params_file_path: Path,
    taxprofiler_sample_sheet_content: str,
    taxprofiler_parameters_default: TaxprofilerParameters,
    caplog: LogCaptureFixture,
):
    """Test that command generates default config files."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a valid case

    # WHEN running config case
    result = cli_runner.invoke(config_case, [taxprofiler_case_id], obj=taxprofiler_context)

    # THEN command should exit successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN logs should be as expected
    expected_logs: list[str] = [
        "Getting sample sheet information",
        "Writing sample sheet",
        "Getting parameters information",
        "Writing parameters file",
    ]
    for expected_log in expected_logs:
        assert expected_log in expected_logs

    # THEN files should be generated
    assert taxprofiler_sample_sheet_path.is_file()
    assert taxprofiler_params_file_path.is_file()

    # THEN the sample sheet content should match the expected values
    sample_sheet_content: list[list[str]] = ReadFile.get_content_from_file(
        file_format=FileFormat.TXT, file_path=taxprofiler_sample_sheet_path, read_to_string=True
    )

    assert ",".join(TaxprofilerSampleSheetEntry.headers()) in sample_sheet_content
    assert taxprofiler_sample_sheet_content in sample_sheet_content

    # THEN the params file should contain all parameters
    params_content: list[list[str]] = ReadFile.get_content_from_file(
        file_format=FileFormat.TXT, file_path=taxprofiler_params_file_path, read_to_string=True
    )
    for parameter in vars(taxprofiler_parameters_default).keys():
        assert parameter in params_content


def test_config_case_dry_run(
    cli_runner: CliRunner,
    taxprofiler_context: CGConfig,
    caplog: LogCaptureFixture,
    taxprofiler_case_id: str,
):
    """Test dry-run."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a valid case

    # WHEN performing a dry-run
    result = cli_runner.invoke(config_case, [taxprofiler_case_id, "-d"], obj=taxprofiler_context)

    # THEN command should exit succesfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN sample sheet and parameters information should be collected
    assert "Getting sample sheet information" in caplog.text
    assert "Getting parameters information" in caplog.text

    # THEN sample sheet and parameters information files should not be written
    assert "Dry run: Config files will not be written" in caplog.text
    assert "Writing sample sheet" not in caplog.text
    assert "Writing parameters file" not in caplog.text



"""This module parses metrics for files generated by the BCLConvert tool using the Dragen hardware."""

import logging
from pathlib import Path
from typing import Callable

from cg.apps.demultiplex.sample_sheet.validators import is_valid_sample_internal_id
from cg.apps.sequencing_metrics_parser.models.bcl_convert import (
    BclConvertDemuxMetrics,
    BclConvertQualityMetrics,
)
from cg.constants.bcl_convert_metrics import (
    ADAPTER_METRICS_FILE_NAME,
    DEMUX_METRICS_FILE_NAME,
    QUALITY_METRICS_FILE_NAME,
)
from cg.constants.constants import SCALE_TO_READ_PAIRS, FileFormat
from cg.constants.demultiplexing import UNDETERMINED
from cg.io.controller import ReadFile
from cg.utils.files import get_file_in_directory

LOG = logging.getLogger(__name__)


class BclConvertMetricsParser:
    def __init__(
        self,
        bcl_convert_metrics_dir_path: Path,
    ) -> None:
        """Initialize the class."""
        self.bcl_convert_demultiplex_dir: Path = bcl_convert_metrics_dir_path
        self.quality_metrics_path: Path = get_file_in_directory(
            directory=self.bcl_convert_demultiplex_dir, file_name=QUALITY_METRICS_FILE_NAME
        )
        self.demux_metrics_path: Path = get_file_in_directory(
            directory=self.bcl_convert_demultiplex_dir, file_name=DEMUX_METRICS_FILE_NAME
        )
        self.adapter_metrics_path: Path = get_file_in_directory(
            directory=self.bcl_convert_demultiplex_dir, file_name=ADAPTER_METRICS_FILE_NAME
        )
        self.quality_metrics: list[BclConvertQualityMetrics] = self.parse_metrics_file(
            metrics_file_path=self.quality_metrics_path,
            metrics_model=BclConvertQualityMetrics,
        )
        self.demux_metrics: list[BclConvertDemuxMetrics] = self.parse_metrics_file(
            metrics_file_path=self.demux_metrics_path,
            metrics_model=BclConvertDemuxMetrics,
        )

    def parse_metrics_file(
        self, metrics_file_path, metrics_model: Callable
    ) -> list[BclConvertQualityMetrics | BclConvertDemuxMetrics]:
        """Parse specified BCL convert metrics file."""
        LOG.info(f"Parsing BCLConvert metrics file: {metrics_file_path}")
        parsed_metrics: list[BclConvertQualityMetrics | BclConvertDemuxMetrics] = []
        metrics_content: list[dict] = ReadFile.get_content_from_file(
            file_format=FileFormat.CSV, file_path=metrics_file_path, read_to_dict=True
        )
        for sample_metrics_dict in metrics_content:
            parsed_metrics.append(metrics_model(**sample_metrics_dict))
        return parsed_metrics

    def get_sample_internal_ids(self) -> list[str]:
        """Return a list of sample internal ids."""
        sample_internal_ids: list[str] = []
        for sample_demux_metric in self.demux_metrics:
            if is_valid_sample_internal_id(
                sample_internal_id=sample_demux_metric.sample_internal_id
            ):
                sample_internal_ids.append(sample_demux_metric.sample_internal_id)
        return list(set(sample_internal_ids))

    def get_lanes_for_sample(self, sample_internal_id: str) -> list[int]:
        """Return a list of lanes for a sample."""
        lanes_for_sample: list[int] = []
        for sample_demux_metric in self.demux_metrics:
            if sample_demux_metric.sample_internal_id == sample_internal_id:
                lanes_for_sample.append(sample_demux_metric.lane)
        return lanes_for_sample

    def get_metrics_for_sample_and_lane(
        self,
        metrics: list[BclConvertQualityMetrics | BclConvertDemuxMetrics],
        sample_internal_id: str,
        lane: int,
    ) -> BclConvertQualityMetrics | BclConvertDemuxMetrics:
        """Return the metrics for a sample by sample internal id."""
        for metric in metrics:
            if metric.sample_internal_id == sample_internal_id and metric.lane == lane:
                return metric

    def get_read_pair_metrics_for_sample_and_lane(
        self, sample_internal_id: str, lane: int
    ) -> list[BclConvertQualityMetrics]:
        """Return the read pair metrics for a sample and lane."""
        read_metrics: list[BclConvertQualityMetrics] = []
        for metric in self.quality_metrics:
            if metric.sample_internal_id == sample_internal_id and metric.lane == lane:
                read_metrics.append(metric)
        return read_metrics

    def calculate_total_reads_for_sample_in_lane(self, sample_internal_id: str, lane: int) -> int:
        """Calculate the total reads for a sample in a lane."""
        metric: BclConvertDemuxMetrics = self.get_metrics_for_sample_and_lane(
            metrics=self.demux_metrics, sample_internal_id=sample_internal_id, lane=lane
        )
        return metric.read_pair_count * SCALE_TO_READ_PAIRS

    def get_q30_bases_percent_for_sample_in_lane(self, sample_internal_id: str, lane: int) -> float:
        """Return the percent of bases that are Q30 for a sample and lane."""
        metrics: list[BclConvertQualityMetrics] = self.get_read_pair_metrics_for_sample_and_lane(
            sample_internal_id=sample_internal_id, lane=lane
        )
        return self.calculate_mean_read_pair_q30_bases_percent(metrics=metrics)

    @classmethod
    def calculate_mean_read_pair_q30_bases_percent(
        cls, metrics: list[BclConvertQualityMetrics]
    ) -> float:
        """Calculate the percent of bases that are Q30 for read pairs."""
        mean_read_pair_q30_bases_percent: float = 0
        for metric in metrics:
            mean_read_pair_q30_bases_percent += metric.q30_bases_percent
        return round(mean_read_pair_q30_bases_percent / SCALE_TO_READ_PAIRS, 2) * 100

    @classmethod
    def calculate_mean_quality_score(cls, metrics: list[BclConvertQualityMetrics]) -> float:
        """Calculate the mean quality score for a list of metrics."""
        total_q_score: float = 0
        for metric in metrics:
            total_q_score += metric.mean_quality_score_q30
        return round(total_q_score / SCALE_TO_READ_PAIRS, 2)

    def get_mean_quality_score_for_sample_in_lane(
        self, sample_internal_id: str, lane: int
    ) -> float:
        """Return the mean quality score for a sample and lane."""
        metrics: list[BclConvertQualityMetrics] = self.get_read_pair_metrics_for_sample_and_lane(
            sample_internal_id=sample_internal_id, lane=lane
        )
        return self.calculate_mean_quality_score(metrics=metrics)

    def has_undetermined_reads_in_lane(self, lane: int) -> bool:
        """Return whether there are undetermined reads in a lane."""
        metrics: BclConvertQualityMetrics | None = self.get_metrics_for_sample_and_lane(
            metrics=self.quality_metrics, sample_internal_id=UNDETERMINED, lane=lane
        )
        return bool(metrics)



"""CLI support to create config and/or start RAREDISEASE."""

import logging

import click

from cg.cli.utils import echo_lines
from cg.cli.workflow.commands import ARGUMENT_CASE_ID, OPTION_DRY
from cg.constants.constants import MetaApis
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.raredisease import RarediseaseAnalysisAPI
from cg.models.cg_config import CGConfig

LOG = logging.getLogger(__name__)


@click.group(invoke_without_command=True)
@click.pass_context
def raredisease(context: click.Context) -> None:
    """NF-core/raredisease analysis workflow."""
    AnalysisAPI.get_help(context)
    context.obj.meta_apis[MetaApis.ANALYSIS_API] = RarediseaseAnalysisAPI(config=context.obj)


@raredisease.command("panel")
@OPTION_DRY
@ARGUMENT_CASE_ID
@click.pass_obj
def panel(context: CGConfig, case_id: str, dry_run: bool) -> None:
    """Write aggregated gene panel file exported from Scout."""

    analysis_api: RarediseaseAnalysisAPI = context.meta_apis["analysis_api"]
    analysis_api.status_db.verify_case_exists(case_internal_id=case_id)

    bed_lines: list[str] = analysis_api.get_gene_panel(case_id=case_id)
    if dry_run:
        echo_lines(lines=bed_lines)
        return
    analysis_api.write_panel(case_id=case_id, content=bed_lines)


@raredisease.command("managed-variants")
@OPTION_DRY
@ARGUMENT_CASE_ID
@click.pass_obj
def managed_variants(context: CGConfig, case_id: str, dry_run: bool) -> None:
    """Write managed variants file exported from Scout."""

    analysis_api: RarediseaseAnalysisAPI = context.meta_apis["analysis_api"]
    analysis_api.status_db.verify_case_exists(case_internal_id=case_id)

    vcf_lines: list[str] = analysis_api.get_managed_variants()
    if dry_run:
        echo_lines(lines=vcf_lines)
        return
    analysis_api.write_managed_variants(case_id=case_id, content=vcf_lines)



import json
import logging
import tempfile
from functools import wraps
from http import HTTPStatus
from pathlib import Path
from typing import Any

import cachecontrol
import requests
from flask import Blueprint, abort, current_app, g, jsonify, make_response, request
from google.auth.transport import requests as google_requests
from google.oauth2 import id_token
from pydantic.v1 import ValidationError
from requests.exceptions import HTTPError
from sqlalchemy.exc import IntegrityError
from urllib3.exceptions import MaxRetryError, NewConnectionError
from werkzeug.utils import secure_filename

from cg.apps.orderform.excel_orderform_parser import ExcelOrderformParser
from cg.apps.orderform.json_orderform_parser import JsonOrderformParser
from cg.constants import ANALYSIS_SOURCES, METAGENOME_SOURCES
from cg.constants.constants import FileFormat
from cg.exc import (
    CaseNotFoundError,
    OrderError,
    OrderFormError,
    OrderNotFoundError,
    TicketCreationError,
)
from cg.io.controller import WriteStream
from cg.meta.orders import OrdersAPI
from cg.models.orders.order import OrderIn, OrderType
from cg.models.orders.orderform_schema import Orderform
from cg.server.dto.delivery_message_response import DeliveryMessageResponse
from cg.server.dto.orders.orders_request import OrdersRequest
from cg.server.dto.orders.orders_response import Order, OrdersResponse
from cg.server.ext import db, lims, osticket
from cg.services.delivery_message.delivery_message_service import DeliveryMessageService
from cg.services.orders.order_service import OrderService
from cg.store.models import (
    Analysis,
    Application,
    ApplicationLimitations,
    Case,
    Customer,
    Flowcell,
    Pool,
    Sample,
    SampleLaneSequencingMetrics,
    User,
)

LOG = logging.getLogger(__name__)
BLUEPRINT = Blueprint("api", __name__, url_prefix="/api/v1")


session = requests.session()
cached_session = cachecontrol.CacheControl(session)


def verify_google_token(token):
    request = google_requests.Request(session=cached_session)
    return id_token.verify_oauth2_token(id_token=token, request=request)


def is_public(route_function):
    @wraps(route_function)
    def public_endpoint(*args, **kwargs):
        return route_function(*args, **kwargs)

    public_endpoint.is_public = True
    return public_endpoint


@BLUEPRINT.before_request
def before_request():
    """Authorize API routes with JSON Web Tokens."""
    if not request.is_secure:
        return abort(
            make_response(jsonify(message="Only https requests accepted"), HTTPStatus.FORBIDDEN)
        )

    if request.method == "OPTIONS":
        return make_response(jsonify(ok=True), HTTPStatus.NO_CONTENT)

    endpoint_func = current_app.view_functions[request.endpoint]
    if getattr(endpoint_func, "is_public", None):
        return

    auth_header = request.headers.get("Authorization")
    if not auth_header:
        return abort(
            make_response(jsonify(message="no JWT token found on request"), HTTPStatus.UNAUTHORIZED)
        )

    jwt_token = auth_header.split("Bearer ")[-1]
    try:
        user_data = verify_google_token(jwt_token)
    except ValueError as e:
        LOG.error(f"Error occurred while decoding JWT token: {e}")
        return abort(
            make_response(jsonify(message="outdated login certificate"), HTTPStatus.UNAUTHORIZED)
        )

    user: User = db.get_user_by_email(user_data["email"])
    if user is None or not user.order_portal_login:
        message = f"{user_data['email']} doesn't have access"
        LOG.error(message)
        return abort(make_response(jsonify(message=message), HTTPStatus.FORBIDDEN))

    g.current_user = user


@BLUEPRINT.route("/submit_order/<order_type>", methods=["POST"])
def submit_order(order_type):
    """Submit an order for samples."""
    api = OrdersAPI(lims=lims, status=db, osticket=osticket)
    error_message: str
    try:
        request_json = request.get_json()
        LOG.info(
            "processing order: %s",
            WriteStream.write_stream_from_content(
                content=request_json, file_format=FileFormat.JSON
            ),
        )
        project = OrderType(order_type)
        order_in = OrderIn.parse_obj(request_json, project=project)

        result = api.submit(
            project=project,
            order_in=order_in,
            user_name=g.current_user.name,
            user_mail=g.current_user.email,
        )
        order_service = OrderService(db)
        order_service.create_order(order_in)
    except (  # user misbehaviour
        OrderError,
        OrderFormError,
        ValidationError,
        ValueError,
    ) as error:
        error_message = error.message if hasattr(error, "message") else str(error)
        http_error_response = HTTPStatus.BAD_REQUEST
        LOG.error(error_message)
    except (  # system misbehaviour
        AttributeError,
        ConnectionError,
        HTTPError,
        IntegrityError,
        KeyError,
        NewConnectionError,
        MaxRetryError,
        TimeoutError,
        TicketCreationError,
        TypeError,
    ) as error:
        LOG.exception(error)
        error_message = error.message if hasattr(error, "message") else str(error)
        http_error_response = HTTPStatus.INTERNAL_SERVER_ERROR
    else:
        return jsonify(
            project=result["project"], records=[record.to_dict() for record in result["records"]]
        )

    if error_message:
        return abort(make_response(jsonify(message=error_message), http_error_response))


@BLUEPRINT.route("/cases")
def get_cases():
    """Return cases with links for a customer from the database."""
    enquiry: str = request.args.get("enquiry")
    action: str = request.args.get("action")

    customers: list[Customer] = _get_current_customers()
    cases: list[Case] = _get_cases(enquiry=enquiry, action=action, customers=customers)

    nr_cases: int = len(cases)
    cases_with_links: list[dict] = [case.to_dict(links=True) for case in cases]
    return jsonify(families=cases_with_links, total=nr_cases)


def _get_current_customers() -> list[Customer] | None:
    """Return customers if the current user is not an admin."""
    return g.current_user.customers if not g.current_user.is_admin else None


def _get_cases(
    enquiry: str | None, action: str | None, customers: list[Customer] | None
) -> list[Case]:
    """Get cases based on the provided filters."""
    return db.get_cases_by_customers_action_and_case_search(
        case_search=enquiry,
        customers=customers,
        action=action,
    )


@BLUEPRINT.route("/cases/<case_id>")
def parse_case(case_id):
    """Return a case with links."""
    case: Case = db.get_case_by_internal_id(internal_id=case_id)
    if case is None:
        return abort(HTTPStatus.NOT_FOUND)
    if not g.current_user.is_admin and (case.customer not in g.current_user.customers):
        return abort(HTTPStatus.FORBIDDEN)
    return jsonify(**case.to_dict(links=True, analyses=True))


@BLUEPRINT.route("/cases/<case_id>/delivery_message", methods=["GET"])
def get_case_delivery_message(case_id: str):
    service = DeliveryMessageService(db)
    try:
        response: DeliveryMessageResponse = service.get_delivery_message(case_id)
        return jsonify(response.model_dump()), HTTPStatus.OK
    except CaseNotFoundError as error:
        return jsonify({"error": str(error)}), HTTPStatus.BAD_REQUEST


@BLUEPRINT.route("/families_in_collaboration")
def parse_families_in_collaboration():
    """Return cases in collaboration."""

    customer_internal_id = request.args.get("customer")
    workflow = request.args.get("data_analysis")
    case_search_pattern = request.args.get("enquiry")

    customer = db.get_customer_by_internal_id(customer_internal_id=customer_internal_id)

    cases = db.get_cases_by_customer_workflow_and_case_search(
        customer=customer, workflow=workflow, case_search=case_search_pattern
    )

    case_dicts = [case.to_dict(links=True) for case in cases]
    return jsonify(families=case_dicts, total=len(cases))


@BLUEPRINT.route("/families_in_collaboration/<family_id>")
def parse_family_in_collaboration(family_id):
    """Return a family with links."""
    case: Case = db.get_case_by_internal_id(internal_id=family_id)
    customer: Customer = db.get_customer_by_internal_id(
        customer_internal_id=request.args.get("customer")
    )
    if case.customer not in customer.collaborators:
        return abort(HTTPStatus.FORBIDDEN)
    return jsonify(**case.to_dict(links=True, analyses=True))


@BLUEPRINT.route("/samples")
def parse_samples():
    """Return samples."""
    if request.args.get("status") and not g.current_user.is_admin:
        return abort(HTTPStatus.FORBIDDEN)
    if request.args.get("status") == "incoming":
        samples: list[Sample] = db.get_samples_to_receive()
    elif request.args.get("status") == "labprep":
        samples: list[Sample] = db.get_samples_to_prepare()
    elif request.args.get("status") == "sequencing":
        samples: list[Sample] = db.get_samples_to_sequence()
    else:
        customers: list[Customer] | None = (
            None if g.current_user.is_admin else g.current_user.customers
        )
        samples: list[Sample] = db.get_samples_by_customer_id_and_pattern(
            pattern=request.args.get("enquiry"), customers=customers
        )
    limit = int(request.args.get("limit", 50))
    parsed_samples: list[dict] = [sample.to_dict() for sample in samples[:limit]]
    return jsonify(samples=parsed_samples, total=len(samples))


@BLUEPRINT.route("/samples_in_collaboration")
def parse_samples_in_collaboration():
    """Return samples in a customer group."""
    customer: Customer = db.get_customer_by_internal_id(
        customer_internal_id=request.args.get("customer")
    )
    samples: list[Sample] = db.get_samples_by_customer_id_and_pattern(
        pattern=request.args.get("enquiry"), customers=customer.collaborators
    )
    limit = int(request.args.get("limit", 50))
    parsed_samples: list[dict] = [sample.to_dict() for sample in samples[:limit]]
    return jsonify(samples=parsed_samples, total=len(samples))


@BLUEPRINT.route("/samples/<sample_id>")
def parse_sample(sample_id):
    """Return a single sample."""
    sample: Sample = db.get_sample_by_internal_id(sample_id)
    if sample is None:
        return abort(HTTPStatus.NOT_FOUND)
    if not g.current_user.is_admin and (sample.customer not in g.current_user.customers):
        return abort(HTTPStatus.FORBIDDEN)
    return jsonify(**sample.to_dict(links=True, flowcells=True))


@BLUEPRINT.route("/samples_in_collaboration/<sample_id>")
def parse_sample_in_collaboration(sample_id):
    """Return a single sample."""
    sample: Sample = db.get_sample_by_internal_id(sample_id)
    customer: Customer = db.get_customer_by_internal_id(
        customer_internal_id=request.args.get("customer")
    )
    if sample.customer not in customer.collaborators:
        return abort(HTTPStatus.FORBIDDEN)
    return jsonify(**sample.to_dict(links=True, flowcells=True))


@BLUEPRINT.route("/pools")
def parse_pools():
    """Return pools."""
    customers: list[Customer] | None = (
        g.current_user.customers if not g.current_user.is_admin else None
    )
    pools: list[Pool] = db.get_pools_to_render(
        customers=customers, enquiry=request.args.get("enquiry")
    )
    parsed_pools: list[dict] = [pool_obj.to_dict() for pool_obj in pools[:30]]
    return jsonify(pools=parsed_pools, total=len(pools))


@BLUEPRINT.route("/pools/<pool_id>")
def parse_pool(pool_id):
    """Return a single pool."""
    pool: Pool = db.get_pool_by_entry_id(entry_id=pool_id)
    if pool is None:
        return abort(HTTPStatus.NOT_FOUND)
    if not g.current_user.is_admin and (pool.customer not in g.current_user.customers):
        return abort(HTTPStatus.FORBIDDEN)
    return jsonify(**pool.to_dict())


@BLUEPRINT.route("/flowcells")
def parse_flow_cells() -> Any:
    """Return flow cells."""
    flow_cells: list[Flowcell] = db.get_flow_cell_by_name_pattern_and_status(
        flow_cell_statuses=[request.args.get("status")],
        name_pattern=request.args.get("enquiry"),
    )
    parsed_flow_cells: list[dict] = [flow_cell.to_dict() for flow_cell in flow_cells[:50]]
    return jsonify(flowcells=parsed_flow_cells, total=len(flow_cells))


@BLUEPRINT.route("/flowcells/<flowcell_id>")
def parse_flow_cell(flowcell_id):
    """Return a single flowcell."""
    flow_cell: Flowcell = db.get_flow_cell_by_name(flow_cell_name=flowcell_id)
    if flow_cell is None:
        return abort(HTTPStatus.NOT_FOUND)
    return jsonify(**flow_cell.to_dict(samples=True))


@BLUEPRINT.route("/flowcells/<flow_cell_name>/sequencing_metrics", methods=["GET"])
def get_sequencing_metrics(flow_cell_name: str):
    """Return sample lane sequencing metrics for a flow cell."""

    if not flow_cell_name:
        return jsonify({"error": "Invalid or missing flow cell id"}), HTTPStatus.BAD_REQUEST

    sequencing_metrics: list[SampleLaneSequencingMetrics] = (
        db.get_sample_lane_sequencing_metrics_by_flow_cell_name(flow_cell_name)
    )

    if not sequencing_metrics:
        return (
            jsonify({"error": f"Sequencing metrics not found for flow cell {flow_cell_name}."}),
            HTTPStatus.NOT_FOUND,
        )

    return jsonify([metric.to_dict() for metric in sequencing_metrics])


@BLUEPRINT.route("/analyses")
def parse_analyses():
    """Return analyses."""
    if request.args.get("status") == "delivery":
        analyses: list[Analysis] = db.get_analyses_to_deliver_for_pipeline()
    elif request.args.get("status") == "upload":
        analyses: list[Analysis] = db.get_analyses_to_upload()
    else:
        analyses: list[Analysis] = db.get_analyses()
    parsed_analysis: list[dict] = [analysis_obj.to_dict() for analysis_obj in analyses[:30]]
    return jsonify(analyses=parsed_analysis, total=len(analyses))


@BLUEPRINT.route("/options")
def parse_options():
    """Return various options."""
    customers: list[Customer | None] = (
        db.get_customers() if g.current_user.is_admin else g.current_user.customers
    )

    app_tag_groups: dict[str, list[str]] = {"ext": []}
    applications: list[Application] = db.get_applications_is_not_archived()
    for application in applications:
        if not application.versions:
            LOG.debug(f"Skipping application {application} that doesn't have a price")
            continue
        if application.is_external:
            app_tag_groups["ext"].append(application.tag)
        if application.prep_category not in app_tag_groups:
            app_tag_groups[application.prep_category]: list[str] = []
        app_tag_groups[application.prep_category].append(application.tag)

    source_groups = {"metagenome": METAGENOME_SOURCES, "analysis": ANALYSIS_SOURCES}

    return jsonify(
        applications=app_tag_groups,
        beds=[bed.name for bed in db.get_active_beds()],
        customers=[
            {
                "text": f"{customer.name} ({customer.internal_id})",
                "value": customer.internal_id,
                "isTrusted": customer.is_trusted,
            }
            for customer in customers
        ],
        organisms=[
            {
                "name": organism.name,
                "reference_genome": organism.reference_genome,
                "internal_id": organism.internal_id,
                "verified": organism.verified,
            }
            for organism in db.get_all_organisms()
        ],
        panels=[panel.abbrev for panel in db.get_panels()],
        sources=source_groups,
    )


@BLUEPRINT.route("/me")
def parse_current_user_information():
    """Return information about current user."""
    if not g.current_user.is_admin and not g.current_user.customers:
        LOG.error(
            "%s is not admin and is not connected to any customers, aborting", g.current_user.email
        )
        return abort(HTTPStatus.FORBIDDEN)

    return jsonify(user=g.current_user.to_dict())


@BLUEPRINT.route("/applications")
@is_public
def parse_applications():
    """Return application tags."""
    applications: list[Application] = db.get_applications_is_not_archived()
    parsed_applications: list[dict] = [application.to_dict() for application in applications]
    return jsonify(applications=parsed_applications)


@BLUEPRINT.route("/applications/<tag>")
@is_public
def parse_application(tag: str):
    """Return an application tag."""
    application: Application = db.get_application_by_tag(tag=tag)
    if not application:
        return abort(make_response(jsonify(message="Application not found"), HTTPStatus.NOT_FOUND))
    return jsonify(**application.to_dict())


@BLUEPRINT.route("/applications/<tag>/pipeline_limitations")
@is_public
def get_application_pipeline_limitations(tag: str):
    """Return application pipeline specific limitations."""
    application_limitations: list[ApplicationLimitations] = db.get_application_limitations_by_tag(
        tag
    )
    if not application_limitations:
        return jsonify(message="Application limitations not found"), HTTPStatus.NOT_FOUND
    return jsonify([limitation.to_dict() for limitation in application_limitations])


@BLUEPRINT.route("/orders")
def get_orders():
    """Return the latest orders."""
    orders_request: OrdersRequest = OrdersRequest.model_validate(request.args.to_dict())
    order_service = OrderService(db)
    response: OrdersResponse = order_service.get_orders(orders_request)
    return make_response(response.model_dump())


@BLUEPRINT.route("/orders/<order_id>")
def get_order(order_id: int):
    """Return an order."""
    order_service = OrderService(db)
    try:
        response: Order = order_service.get_order(order_id)
        response_dict: dict = response.model_dump()
        return make_response(response_dict)
    except OrderNotFoundError as error:
        return make_response(jsonify(error=str(error)), HTTPStatus.NOT_FOUND)


@BLUEPRINT.route("/orderform", methods=["POST"])
def parse_orderform():
    """Parse an orderform/JSON export."""
    input_file = request.files.get("file")
    filename = secure_filename(input_file.filename)

    error_message: str
    try:
        if filename.lower().endswith(".xlsx"):
            temp_dir = Path(tempfile.gettempdir())
            saved_path = str(temp_dir / filename)
            input_file.save(saved_path)
            order_parser = ExcelOrderformParser()
            order_parser.parse_orderform(excel_path=saved_path)
        else:
            json_data = json.load(input_file.stream, strict=False)
            order_parser = JsonOrderformParser()
            order_parser.parse_orderform(order_data=json_data)
        parsed_order: Orderform = order_parser.generate_orderform()
    except (  # user misbehaviour
        AttributeError,
        OrderFormError,
        OverflowError,
        ValidationError,
        ValueError,
    ) as error:
        error_message = error.message if hasattr(error, "message") else str(error)
        LOG.error(error_message)
        http_error_response = HTTPStatus.BAD_REQUEST
    except (  # system misbehaviour
        NewConnectionError,
        MaxRetryError,
        TimeoutError,
        TypeError,
    ) as error:
        LOG.exception(error)
        error_message = error.message if hasattr(error, "message") else str(error)
        http_error_response = HTTPStatus.INTERNAL_SERVER_ERROR
    else:
        return jsonify(**parsed_order.model_dump())

    if error_message:
        return abort(make_response(jsonify(message=error_message), http_error_response))



from typing import Any

from sqlalchemy.orm import Query

from cg.constants.constants import SampleType
from cg.constants.subject import PhenotypeStatus, Sex
from cg.store.filters.status_sample_filters import (
    filter_samples_by_entry_customer_ids,
    filter_samples_by_entry_id,
    filter_samples_by_identifier_name_and_value,
    filter_samples_by_internal_id,
    filter_samples_by_internal_id_pattern,
    filter_samples_by_invoice_id,
    filter_samples_by_name,
    filter_samples_by_subject_id,
    filter_samples_do_invoice,
    filter_samples_is_delivered,
    filter_samples_is_not_delivered,
    filter_samples_is_not_down_sampled,
    filter_samples_is_not_prepared,
    filter_samples_is_not_received,
    filter_samples_is_not_sequenced,
    filter_samples_is_prepared,
    filter_samples_is_received,
    filter_samples_is_sequenced,
    filter_samples_with_loqusdb_id,
    filter_samples_with_type,
    filter_samples_without_invoice_id,
    filter_samples_without_loqusdb_id,
)
from cg.store.models import CaseSample, Sample
from cg.store.store import Store
from tests.store.conftest import StoreConstants
from tests.store_helpers import StoreHelpers


def test_get_samples_with_loqusdb_id(helpers, store, sample_store, sample_id, loqusdb_id):
    """Test sample extraction with Loqusdb ID."""

    # GIVEN a sample observations that has been uploaded to Loqusdb
    case = helpers.add_case(store)
    sample = helpers.add_sample(store, loqusdb_id=loqusdb_id)
    sample_not_uploaded = helpers.add_sample(store, internal_id=sample_id)
    link_1: CaseSample = sample_store.relate_sample(
        case=case, sample=sample, status=PhenotypeStatus.UNKNOWN
    )
    link_2: CaseSample = sample_store.relate_sample(
        case=case, sample=sample_not_uploaded, status=PhenotypeStatus.UNKNOWN
    )
    sample_store.session.add_all([link_1, link_2])

    # GIVEN a sample query
    samples: Query = store._get_query(table=Sample)

    # WHEN retrieving the Loqusdb uploaded samples
    uploaded_samples = filter_samples_with_loqusdb_id(samples=samples)

    # THEN the obtained sample should match the expected one
    assert sample in uploaded_samples
    assert sample_not_uploaded not in uploaded_samples


def test_get_samples_without_loqusdb_id(
    helpers: StoreHelpers, sample_store: Store, sample_id, loqusdb_id
):
    """Test sample extraction without Loqusdb ID."""

    # GIVEN a sample observations that has not been uploaded to Loqusdb
    sample = helpers.add_sample(sample_store)
    sample_uploaded = helpers.add_sample(
        store=sample_store, internal_id=sample_id, loqusdb_id=loqusdb_id
    )

    # GIVEN a sample query
    samples: Query = sample_store._get_query(table=Sample)

    # WHEN retrieving the Loqusdb not uploaded samples
    not_uploaded_samples = filter_samples_without_loqusdb_id(samples)

    # THEN the obtained sample should match the expected one
    assert sample in not_uploaded_samples
    assert sample_uploaded not in not_uploaded_samples


def test_filter_samples_is_delivered(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
):
    """Test that a sample is returned when there is a delivered sample."""

    # GIVEN a store with two samples of which one is delivered

    # WHEN getting delivered samples
    samples: Query = filter_samples_is_delivered(
        samples=store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
            table=Sample
        )
    )

    # ASSERT that samples is a query
    assert isinstance(samples, Query)

    # THEN samples should contain the test sample
    assert samples.all()

    # THEN samples should contain one sample
    assert len(samples.all()) == 1

    # THEN the sample should have a delivered at date
    assert samples.all()[0].delivered_at is not None


def test_filter_samples_is_not_delivered(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
):
    """Test that a sample is returned when there is a sample that is not delivered."""

    # GIVEN a store with two samples of which one is not delivered

    # WHEN getting not sequenced samples
    samples: Query = filter_samples_is_not_delivered(
        samples=store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
            table=Sample
        )
    )

    # ASSERT that samples is a query
    assert isinstance(samples, Query)

    # THEN samples should contain the test sample
    assert samples.all()

    # THEN samples should contain one sample
    assert len(samples.all()) == 1

    # THEN the sample should not have a delivered at date
    assert samples.all()[0].delivered_at is None


def test_filter_get_samples_by_invoice_id(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
    invoice_id=StoreConstants.INVOICE_ID_SAMPLE_WITH_ATTRIBUTES.value,
):
    """Test that a sample is returned when there is a sample that has an invoice id."""

    # GIVEN a store with two samples of which one has an invoice id

    # WHEN getting not sequenced samples
    samples: Query = filter_samples_by_invoice_id(
        samples=store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
            table=Sample
        ),
        invoice_id=invoice_id,
    )

    # ASSERT that samples is a query
    assert isinstance(samples, Query)

    # THEN samples should contain the test sample
    assert samples.all()

    # THEN samples should contain one sample
    assert len(samples.all()) == 1

    # THEN the sample should have the correct invoice id
    assert samples.all()[0].invoice_id == invoice_id


def test_filter_samples_without_invoice_id(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
):
    """Test that a sample is returned when there is a sample that has no invoice id."""

    # GIVEN a store with two samples of which one does not have an invoice id

    # WHEN getting not sequenced samples
    samples: Query = filter_samples_without_invoice_id(
        samples=store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
            table=Sample
        )
    )

    # ASSERT that samples is a query
    assert isinstance(samples, Query)

    # THEN samples should contain the test sample
    assert samples.all()

    # THEN samples should contain one sample
    assert len(samples.all()) == 1

    # THEN the sample should not have an invoice id
    assert samples.all()[0].invoice_id is None


def test_filter_samples_not_down_sampled(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
):
    """Test that a sample is returned when there is a sample that is not down sampled."""

    # GIVEN a store with two samples of which one is not sequenced

    # WHEN getting not sequenced samples
    samples: Query = filter_samples_is_not_down_sampled(
        samples=store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
            table=Sample
        )
    )

    # ASSERT that samples is a query
    assert isinstance(samples, Query)

    # THEN samples should contain the test sample
    assert samples.all()

    # THEN samples should contain one sample
    assert len(samples.all()) == 1

    # THEN the sample should not have a down sampled to value
    assert samples.all()[0].downsampled_to is None


def test_filter_samples_is_sequenced(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
):
    """Test that a sample is returned when there is a sample that is not sequenced."""

    # GIVEN a store with two samples of which one is not sequenced

    # WHEN getting not sequenced samples
    samples: Query = filter_samples_is_sequenced(
        samples=store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
            table=Sample
        )
    )

    # ASSERT that samples is a query
    assert isinstance(samples, Query)

    # THEN samples should contain the test sample
    assert samples.all()

    # THEN samples should contain one sample
    assert len(samples.all()) == 1

    # THEN the sample should have a sequenced at date
    assert samples.all()[0].last_sequenced_at is not None


def test_filter_samples_is_not_sequenced(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
):
    """Test that a sample is returned when there is a sample that is not sequenced."""

    # GIVEN a store with two samples of which one is not sequenced

    # WHEN getting not sequenced samples
    samples: Query = filter_samples_is_not_sequenced(
        samples=store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
            table=Sample
        )
    )

    # ASSERT that samples is a query
    assert isinstance(samples, Query)

    # THEN samples should contain the test sample
    assert samples.all()

    # THEN samples should contain one sample
    assert len(samples.all()) == 1

    # THEN the sample should not have a sequenced at date
    assert samples.all()[0].last_sequenced_at is None


def test_filter_samples_do_invoice(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
):
    """Test that a sample is returned when there is not a sample that should be invoiced."""

    # GIVEN a  store with two samples of which one is marked to skip invoicing

    # WHEN getting  samples mark to be invoiced
    samples: Query = filter_samples_do_invoice(
        samples=store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
            table=Sample
        )
    )

    # ASSERT that samples is a query
    assert isinstance(samples, Query)

    # THEN samples should contain the test sample
    assert samples.all()

    # THEN samples should contain one sample
    assert len(samples.all()) == 1

    # THEN the sample should have a no invoice indicator that is set to False
    assert samples.all()[0].no_invoice is False


def test_filter_samples_is_received(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
):
    """Test that a sample is returned when there is a sample that is received."""

    # GIVEN a store with two samples of which one is received

    # WHEN getting received samples
    samples: Query = filter_samples_is_received(
        samples=store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
            table=Sample
        )
    )

    # ASSERT that samples is a query
    assert isinstance(samples, Query)

    # THEN samples should contain the test sample
    assert samples.all()

    # THEN samples should contain one sample
    assert len(samples.all()) == 1

    # THEN the sample should have a received at date
    assert samples.all()[0].received_at is not None


def test_filter_samples_is_not_received(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
):
    """Test that a sample is returned when there is a sample that is not received."""

    # GIVEN a store that has two samples of which one is not received

    # WHEN getting not received samples
    samples: Query = filter_samples_is_not_received(
        samples=store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
            table=Sample
        )
    )

    # THEN samples should contain the test sample
    assert samples.all()

    # THEN samples should contain one sample
    assert len(samples.all()) == 1

    # THEN the sample should not have a received at date
    assert samples.all()[0].received_at is None


def test_filter_samples_is_prepared(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
):
    """Test that a sample is returned when there is a sample that is prepared."""

    # GIVEN a store that has two samples of which one is prepared

    samples: Query = filter_samples_is_prepared(
        samples=store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
            table=Sample
        )
    )

    # ASSERT that samples is a query
    assert isinstance(samples, Query)

    # THEN samples should contain the test sample
    assert samples.all()

    # THEN samples should contain one sample
    assert len(samples.all()) == 1

    # THEN the sample should have a prepared at date
    assert samples.all()[0].prepared_at is not None


def test_filter_samples_is_not_prepared(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
):
    """Test that a sample is returned when there is a sample that is not prepared."""

    # GIVEN a store that has two samples of which one is not prepared

    # WHEN getting not prepared samples
    samples: Query = filter_samples_is_not_prepared(
        samples=store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
            table=Sample
        )
    )

    # THEN samples should contain the test sample
    assert samples.all()

    # THEN samples should contain one sample
    assert len(samples.all()) == 1

    # THEN the sample should have not have a prepared at date
    assert samples.all()[0].prepared_at is None


def test_filter_get_samples_by_internal_id(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
    sample_internal_id: str = StoreConstants.INTERNAL_ID_SAMPLE_WITH_ATTRIBUTES.value,
):
    """Test that a sample is returned when there is a sample with the given id."""

    # GIVEN a store with two samples of which one has the given internal id

    # WHEN getting a sample by id
    samples: Query = filter_samples_by_internal_id(
        samples=store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
            table=Sample
        ),
        internal_id=sample_internal_id,
    )

    # ASSERT that samples is a query
    assert isinstance(samples, Query)

    # THEN samples should contain the test sample
    assert samples.all()

    # THEN samples should contain one sample
    assert len(samples.all()) == 1

    # THEN the sample should have the internal id
    assert samples.all()[0].internal_id == sample_internal_id


def test_filter_get_samples_by_entry_id(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
    entry_id: int = 1,
):
    """Test that a sample is returned when there is a sample with the given entry id."""

    # GIVEN a store with two samples

    # WHEN getting a sample by id
    samples: Query = filter_samples_by_entry_id(
        samples=store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
            table=Sample
        ),
        entry_id=entry_id,
    )

    # ASSERT that samples is a query
    assert isinstance(samples, Query)

    # THEN samples should contain the test sample
    assert samples.all()

    # THEN samples should contain one sample
    assert len(samples.all()) == 1

    # THEN the sample should have the correct id
    assert samples.all()[0].id == entry_id


def test_filter_get_samples_with_type(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
    name=StoreConstants.NAME_SAMPLE_WITH_ATTRIBUTES.value,
    tissue_type: SampleType = SampleType.TUMOR,
):
    """Test that a sample is returned when there is a sample with the given type."""

    # GIVEN a store with two samples of which one is of the given type

    # WHEN getting a sample by type
    samples: Query = filter_samples_with_type(
        samples=store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
            table=Sample
        ),
        tissue_type=tissue_type,
    )

    # ASSERT that samples is a query
    assert isinstance(samples, Query)

    # THEN samples should contain the test sample
    assert samples.all()

    # THEN samples should contain one sample
    assert len(samples.all()) == 1

    # THEN the sample should have is tumour set to true
    assert samples.all()[0].is_tumour is True


def test_filter_get_samples_by_name(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
    name=StoreConstants.NAME_SAMPLE_WITH_ATTRIBUTES.value,
):
    """Test that a sample is returned when there is a sample with the given type."""
    # GIVEN a store with two samples that have names

    # WHEN getting a sample by name
    samples: Query = filter_samples_by_name(
        samples=store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
            table=Sample
        ),
        name=name,
    )

    # ASSERT that samples is a query
    assert isinstance(samples, Query)

    # THEN samples should contain the test sample
    assert samples.all()

    # THEN samples should contain one sample
    assert len(samples.all()) == 1

    # THEN the sample should have the correct name
    assert samples.all()[0].name == name


def test_filter_get_samples_by_subject_id(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
    subject_id: str = StoreConstants.SUBJECT_ID_SAMPLE_WITH_ATTRIBUTES.value,
):
    """Test that a sample is returned when there is a sample with the given subject id."""
    # GIVEN a store with two samples of which one has a subject id

    # WHEN getting a sample by subject id
    samples: Query = filter_samples_by_subject_id(
        samples=store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
            table=Sample
        ),
        subject_id=subject_id,
    )

    # ASSERT that samples is a query
    assert isinstance(samples, Query)

    # THEN samples should contain the test sample
    assert samples.all()

    # THEN samples should contain one sample
    assert len(samples.all()) == 1

    # THEN the sample should have the correct subject id
    assert samples.all()[0].subject_id == subject_id


def test_filter_get_samples_by_customer_id(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
):
    """Test that a sample is returned when there is a sample with the given customer id."""
    # GIVEN a store with samples with different customer ids
    samples: Query = store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
        table=Sample
    )
    customer_id: int = samples.first().customer_id
    assert customer_id != samples.order_by(Sample.customer_id.desc()).first().customer_id

    # WHEN filtering the sample query by customer id
    filtered_query: Query = filter_samples_by_entry_customer_ids(
        samples=samples,
        customer_entry_ids=[customer_id],
    )

    # THEN the result of the filtering is a query
    assert isinstance(filtered_query, Query)

    # THEN the filtered query is not empty
    assert filtered_query.all()

    # THEN the filtered query has fewer elements than the unfiltered query
    assert filtered_query.count() < samples.count()

    # THEN a sample in the filtered query should have the correct customer id
    assert filtered_query.first().customer_id == customer_id


def test_filter_get_samples_by_internal_id_pattern(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
    internal_id_pattern: str = "with_attributes",
):
    """Test that a sample is returned when there is a sample with the given internal id pattern."""
    # GIVEN a store with two samples of which one has a name name pattern

    # WHEN getting a sample by name pattern
    samples: Query = filter_samples_by_internal_id_pattern(
        samples=store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
            table=Sample
        ),
        internal_id_pattern=internal_id_pattern,
    )

    # ASSERT that samples is a query
    assert isinstance(samples, Query)

    # THEN samples should contain the test sample
    assert samples.all()

    # THEN samples should contain one sample
    assert len(samples.all()) == 1

    # THEN the sample should have the correct name
    assert samples[0].internal_id == StoreConstants.INTERNAL_ID_SAMPLE_WITH_ATTRIBUTES.value


def test_filter_samples_by_identifier_name_and_value_unique_sample(
    store_with_a_sample_that_has_many_attributes_and_one_without: Store,
):
    """Test that the function filters correctly for any identifier."""
    # GIVEN a store with at least two samples
    sample_query: Query = store_with_a_sample_that_has_many_attributes_and_one_without._get_query(
        table=Sample
    )
    assert sample_query.count() > 1

    # GIVEN a sample in store that has all attributes
    sample: Sample = sample_query.first()

    # WHEN filtering the sample query with every existing attribute of the sample
    identifiers: dict[str, Any] = {
        "age_at_sampling": sample.age_at_sampling,
        "application_version_id": sample.application_version_id,
        "capture_kit": sample.capture_kit,
        "comment": sample.comment,
        "control": sample.control,
        "created_at": sample.created_at,
        "customer_id": sample.customer_id,
        "delivered_at": sample.delivered_at,
        "downsampled_to": sample.downsampled_to,
        "from_sample": sample.from_sample,
        "id": sample.id,
        "internal_id": sample.internal_id,
        "invoice_id": sample.invoice_id,
        "invoiced_at": sample.invoiced_at,
        "is_tumour": sample.is_tumour,
        "loqusdb_id": sample.loqusdb_id,
        "name": sample.name,
        "no_invoice": sample.no_invoice,
        "order": sample.order,
        "ordered_at": sample.ordered_at,
        "organism_id": sample.organism_id,
        "original_ticket": sample.original_ticket,
        "prepared_at": sample.prepared_at,
        "priority": sample.priority,
        "reads": sample.reads,
        "received_at": sample.received_at,
        "reference_genome": sample.reference_genome,
        "sequence_start": sample.sequence_start,
        "sex": sample.sex,
        "last_sequenced_at": sample.last_sequenced_at,
        "subject_id": sample.subject_id,
    }
    for key, value in identifiers.items():
        filtered_sample_query: Query = filter_samples_by_identifier_name_and_value(
            samples=sample_query,
            identifier_name=key,
            identifier_value=value,
        )
        # THEN the filtered query has at least one element
        assert filtered_sample_query.count() > 0
        # THEN the element in the filtered query is the sample for every attribute
        assert getattr(filtered_sample_query.first(), key) == value


def test_filter_samples_by_identifier_name_and_value_two_samples(sample_store: Store):
    """."""
    # GIVEN a store with more than 2 samples
    sample_query: Query = sample_store._get_query(table=Sample)
    assert sample_query.count() > 2

    # WHEN filtering the females from the sample query using identifiers
    filtered_query: Query = filter_samples_by_identifier_name_and_value(
        samples=sample_query,
        identifier_name="sex",
        identifier_value=Sex.FEMALE,
    )

    # THEN the filtered query has at least two elements
    assert filtered_query.count() > 1

    # THEN all the elements of the filtered query are females
    for sample in filtered_query:
        assert sample.sex == Sex.FEMALE
